#! /usr/bin/env python
#
# Mixpanel, Inc. -- http://mixpanel.com/
#
# Python API client library to consume mixpanel.com analytics data.

import hashlib
import urllib
import time
try:
    import json
except ImportError:
    import simplejson as json

class Mixpanel(object):

    ENDPOINT = 'http://mixpanel.com/api'
    VERSION = '2.0'

    def __init__(self, api_key, api_secret):
        self.api_key = api_key
        self.api_secret = api_secret
        
    def request(self, methods, params, format='json'):
        """
            methods - List of methods to be joined, e.g. ['events', 'properties', 'values']
                      will give us http://mixpanel.com/api/2.0/events/properties/values/
            params - Extra parameters associated with method
        """
        params['api_key'] = self.api_key
        params['expire'] = int(time.time()) + 600   # Grant this request 10 minutes.
        params['format'] = format
        if 'sig' in params: del params['sig']
        params['sig'] = self.hash_args(params)

        request_url = '/'.join([self.ENDPOINT, str(self.VERSION)] + methods) + '/?' + self.unicode_urlencode(params)

        request = urllib.urlopen(request_url)
        data = request.read()

        return str(data)

    def unicode_urlencode(self, params):
        """
            Convert lists to JSON encoded strings, and correctly handle any 
            unicode URL parameters.
        """
        if isinstance(params, dict):
            params = params.items()
        for i, param in enumerate(params):
            if isinstance(param[1], list): 
                params[i] = (param[0], json.dumps(param[1]),)

        return urllib.urlencode(
            [(k, isinstance(v, unicode) and v.encode('utf-8') or v) for k, v in params]
        )

    def hash_args(self, args, secret=None):
        """
            Hashes arguments by joining key=value pairs, appending a secret, and 
            then taking the MD5 hex digest.
        """
        for a in args:
            if isinstance(args[a], list): args[a] = json.dumps(args[a])

        args_joined = ''
        for a in sorted(args.keys()):
            if isinstance(a, unicode):
                args_joined += a.encode('utf-8')
            else:
                args_joined += str(a)

            args_joined += '='

            if isinstance(args[a], unicode):
                args_joined += args[a].encode('utf-8')
            else:
                args_joined += str(args[a])

        hash = hashlib.md5(args_joined)

        if secret:
            hash.update(secret)
        elif self.api_secret:
            hash.update(self.api_secret)
        return hash.hexdigest() 

if __name__ == '__main__':
    api = Mixpanel(
        api_key = 'YOUR KEY', 
        api_secret = 'YOUR SECRET'
    )
    data = api.request(['events'], {
        'event' : ['pages',],
        'unit' : 'hour',
        'interval' : 24,
        'type': 'general'
    })
    print data



# -*- coding: utf-8 -*-

'''

    importer: timezone tests
    ~~~~~~~~~~~~~~~~~~~~~~~~

    this module tests functionality in ``importer``
    related to timezone and time offset computation.

    :author: Sam Gammon <sam@keen.io>
    :license: This software follows the MIT (OSI-approved)
              license for open source software. A truncated
              version is included here; for full licensing
              details, see ``LICENSE.md`` in the root directory
              of the project.

              Copyright (c) 2013, Keen IO

              The above copyright notice and this permission notice shall be included in
              all copies or substantial portions of the Software.

              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
              IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
              FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
              AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
              LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
              OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
              THE SOFTWARE.

'''


# stdlib
import time
import unittest
import datetime

# importer
from importer import tz
from importer import mix


## TZTests - puts offset calculation functions through their paces.
class TZTests(unittest.TestCase):

    ''' Tests routines related to calculating
        timezone/DST offsets. '''

    def test_exports(self):

        ''' Make sure that all expected module globals are exported
            where they should be. '''

        self.assertTrue(hasattr(tz, 'timezone'))  # test that proxy exists to :py:mod:`pytz`



# -*- coding: utf-8 -*-

'''

    importer
    ~~~~~~~~

    small utility for transferring evented data from one
    provider to another. full description coming soon! :)

    :author: Sam Gammon <sam@keen.io>
    :license: This software follows the MIT (OSI-approved)
              license for open source software. A truncated
              version is included here; for full licensing
              details, see ``LICENSE.md`` in the root directory
              of the project.

              Copyright (c) 2013, Keen IO

              The above copyright notice and this permission notice shall be included in
              all copies or substantial portions of the Software.

              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
              IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
              FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
              AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
              LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
              OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
              THE SOFTWARE.

'''

__version__ = (1, 1)
__author__ = "Sam Gammon <sam@keen.io>"



# -*- coding: utf-8 -*-

'''
    importer: timezones
    ~~~~~~~~~~~~~~~~~~~

    this package provides utilities and predefined
    dictionaries for dealing with timezones and
    time offsets.

    :author: Sam Gammon <sam@keen.io>
    :license: This software follows the MIT (OSI-approved)
              license for open source software. A truncated
              version is included here; for full licensing
              details, see ``LICENSE.md`` in the root directory
              of the project.

              Copyright (c) 2013, Keen IO

              The above copyright notice and this permission notice shall be included in
              all copies or substantial portions of the Software.

              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
              IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
              FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
              AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
              LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
              OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
              THE SOFTWARE.

'''

__version__ = (1, 1)

# stdlib
import logging

# pytz
try:
    from pytz import timezone

except ImportError as e:  # pragma: nocover
    logging.critical('Failed to load `pytz`, which `Importer` depends on. Exiting.')
    exit(1)


# `_TIMEZONE_DATA`: compound data dictionary of timezone names, offsets, DST's, and keys
_TIMEZONE_DATA = (
    (-12, 'BIT', 'Baker Island Time', 'Pacific/Baker'),
    (-11.5, 'NUT', 'Niue Time', 'Pacific/Niue'),
    (-11, 'SST', 'Samoa Standard Time', 'Pacific/Samoa'),
    (-10, 'CKT', 'Cook Island Time', 'Pacific/Cook_Islands'),
    (-9.5, 'MART', 'Marquesas Islands Time', 'Pacific/Marquesas'),
    (-9, 'AKST', 'Alaska Standard Time', 'US/Alaska'),
    (-8, 'PST', 'Pacific Standard Time', 'US/Pacific'),
    (-7, 'MST', 'Mountain Standard Time', 'US/Mountain'),
    (-6, 'CST', 'Central Standard Time', 'US/Central'),
    (-5, 'EST', 'Eastern Standard Time', 'US/Eastern'),
    (-4.5, 'VET', 'Venezuelan Standard Time', 'Americas/Venezuela'),
    (-4, 'AST', 'Atlantic Standard Time', 'Americas/Atlantic'),
    (-3.5, 'NST', 'Newfoundland Standard Time', 'Americas/Newfoundland'),
    (-3, 'ADT', 'Atlantic Daylight Time', 'Americas/Atlantic/DST'),
    (-2.5, 'NDT', 'Newfoundland Daylight Time', 'Americas/Newfoundland/DST'),
    (-2, 'FNT', 'Fernando de Noronha Time', 'European/Spain'),
    (-1, 'EGT', 'East Greenland Time', 'European/East_Greenland'),
    (0, 'UTC', 'UTC', 'UTC'),
    (1, 'CET', 'Central European Time', 'European/Central'),
    (2, 'CAT', 'Central Africa Time', 'Africa/Central'),
    (3, 'AST', 'Arabia Standard Time', 'MiddleEast/Arabia_Standard'),
    (3.5, 'IRST', 'Iran Standard Time', 'MiddleEast/Iran'),
    (4, 'AZT', 'Azerbaijan Time', 'MiddleEast/Azerbaijan'),
    (4.5, 'AFT', 'Afghanistan Time', 'MiddleEast/Afghanistan'),
    (5, 'MVT', 'Maldives Time', 'Asia/Maldives'),
    (5.5, 'IST', 'Indian Standard Time', 'Asia/India_Standard'),
    (6, 'BST', 'Bangladesh Standard Time', 'Asia/Bangladesh'),
    (6.5, 'CCT', 'Cocos Islands Time', 'Oceania/Cocos'),
    (7, 'CXT', 'Christmas Island Time', 'Oceania/Christmas_Island'),
    (8, 'ACT', 'ASEAN Common Time', 'Asia/Common'),
    (9, 'JST', 'Japan Standard Time', 'Asia/Japan'),
    (9.5, 'ACST', 'Australian Central Standard Time', 'Australia/Central'),
    (10, 'AEST', 'Australian Eastern Standard Time', 'Australia/Eastern'),
    (10.5, 'ACDT', 'Australian Central Daylight Time', 'Australia/Central/DST'),
    (11, 'AEDT', 'Australian Eastern Daylight Time', 'Australia/Eastern/DST'),
    (11.5, 'NFT', 'Norfolk Time', 'Oceania/Norfolk'),
    (12, 'FJT', 'Fiji Time', 'Oceania/Fiji')
)


# `_TZNAMES`: mapping from long (config) names to offsets
_TZNAMES = dict(((name, offset) for offset, shortcode, longname, name in _TIMEZONE_DATA))


# `_TZSHORT`: mapping from timezone offsets to shortcodes
_TZSHORT = dict(((offset, shortcode) for offset, shortcode, longname, name in _TIMEZONE_DATA))


# `_TIMEZONES`: map TZ offsets to full data
_TIMEZONES = dict(((offset, (shortcode, longname, name)) for offset, shortcode, longname, name in _TIMEZONE_DATA))



# -*- coding: utf-8 -*-

'''

    importer: keen tests
    ~~~~~~~~~~~~~~~~~~~~

    this module tests ``Keen``-specific provider functionality.

    :author: Sam Gammon <sam@keen.io>
    :license: This software follows the MIT (OSI-approved)
              license for open source software. A truncated
              version is included here; for full licensing
              details, see ``LICENSE.md`` in the root directory
              of the project.

              Copyright (c) 2013, Keen IO

              The above copyright notice and this permission notice shall be included in
              all copies or substantial portions of the Software.

              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
              IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
              FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
              AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
              LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
              OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
              THE SOFTWARE.

'''


# stdlib
import json
import time
import random
import unittest
import datetime

# importer
from importer import mix

# main keen-related classes
from importer.mix import Keen
from importer.mix import KeenUploader
from importer.mix import KeenDownloader


## KeenTests - tests general integration with Keen.
class KeenTests(unittest.TestCase):

    ''' Tests the ``Keen`` provider class, which is
        exposed at :py:class:`mix.Keen`. '''

    def _valid_keen_provider(self, instance=False, params=False):  # pragma: nocover

        ''' Generate a valid :py:class:`Keen` provider
            object, with minimal config/etc for use
            in testing. '''

        ## `add_events` shim
        def _add_events(self, events):

            ''' Quick shim to simulate a mocked ``Keen``
                library, supporting batch event writes. '''

            self.events_cache = events
            return self

        ## `extraction` shim
        def _extraction(self, kind, range):

            ''' Quick shim to simulate a mocked ``Keen``
                library, supporting event extractions. '''

            # manufacture known datetimes
            dt = datetime.datetime(year=2013, month=8, day=24, hour=12, minute=30, second=0)
            dt2 = datetime.datetime(year=2013, month=8, day=25, hour=12, minute=35, second=0)
            
            return [
                {'kind': kind, 'prop': 123, 'keen': {'timestamp': mix._TO_KEEN_DATE(dt)}},
                {'kind': kind, 'prop': 124, 'keen': {'timestamp': mix._TO_KEEN_DATE(dt2)}}
            ]

        ## prepare args
        _bus_, _name_, _library_, _adapters_, _config_ = (
            object(),  # bus mock
            'KEEN',  # provider name
            type('KeenMock', (object,), {
                'project_id': None,
                'read_key': None,
                'write_key': None,
                'add_events': _add_events,
                'extraction': _extraction,
                'events_cache': None})(),  # library mock
            (KeenUploader, KeenDownloader),  # bound adapters
            {
                'read_key': '__read_key__',
                'write_key': '__write_key__',
                'project_id': '__project_id__'
            }
        )

        if instance:
            k = Keen(_bus_, _name_, _library_, _adapters_, **_config_)
            if params:
                return (_bus_, _name_, _library_, _adapters_, _config_), k
            return k
        if params:
            return (_bus_, _name_, _library_, _adapters_, _config_), Keen
        return Keen

    def test_keen_construct(self):

        ''' Test the :py:class:`Keen` class, which
            implements the :py:class:`Provider`
            interface. '''

        _keen_args, KeenProvider = self._valid_keen_provider(instance=False, params=True)
        _bus, _name, _library, _adapters, _config = _keen_args

        # try to construct without adapters
        with self.assertRaises(TypeError):
            KeenProvider(_bus, _name, _library, tuple(), **_config)

        # try to construct with incomplete adapters
        with self.assertRaises(TypeError):
            KeenProvider(_bus, _name, _library, (KeenUploader,), **_config)

        # try to construct in a valid way
        _valid_keen = KeenProvider(_bus, _name, _library, _adapters, **_config)

        # test attached parameters
        self.assertEqual(_valid_keen.bus, _bus)
        self.assertEqual(_valid_keen.lib, _library)
        self.assertEqual(_valid_keen.name, 'KEEN')
        self.assertIsInstance(_valid_keen.adapters, tuple)
        self.assertEqual(len(_valid_keen.adapters), 2)

    def test_keen_config(self):

        ''' Test :py:attr:`Keen.config`, which exposes
            ``Keen``-related provider config listed in
            *importer's* ``config.json``. '''

        _keen_provider = self._valid_keen_provider(instance=True)

        # check config values sent in by the valid provider util
        self.assertIsInstance(_keen_provider.config, dict)
        self.assertEqual(_keen_provider.config['read_key'], '__read_key__')
        self.assertEqual(_keen_provider.config['write_key'], '__write_key__')
        self.assertEqual(_keen_provider.config['project_id'], '__project_id__')

    def test_keen_key_methods(self):

        ''' Test utility methods exposed on :py:class:`Keen`
            that allow direct API users to change/set
            their config values (``project_id``, ``read_key``,
            and ``write_key``) in-flight. '''

        _keen_provider = self._valid_keen_provider(instance=True)

        # check getter methods, implicitly
        self.assertEqual(_keen_provider.read_key, _keen_provider.config['read_key'])
        self.assertEqual(_keen_provider.write_key, _keen_provider.config['write_key'])
        self.assertEqual(_keen_provider.project_id, _keen_provider.config['project_id'])

        # check getter methods, explicitly
        self.assertEqual(_keen_provider._get_read_key(), _keen_provider.config['read_key'])
        self.assertEqual(_keen_provider._get_write_key(), _keen_provider.config['write_key'])
        self.assertEqual(_keen_provider._get_project_id(), _keen_provider.config['project_id'])

        # check setter methods, implicitly
        _keen_provider.project_id = 'supguyz'
        self.assertEqual(_keen_provider.project_id, 'supguyz')
        self.assertEqual(_keen_provider.config['project_id'], 'supguyz')

        _keen_provider.read_key = 'readkey_sample'
        self.assertEqual(_keen_provider.read_key, 'readkey_sample')
        self.assertEqual(_keen_provider.config['read_key'], 'readkey_sample')

        _keen_provider.write_key = 'writekey_sample'
        self.assertEqual(_keen_provider.write_key, 'writekey_sample')
        self.assertEqual(_keen_provider.config['write_key'], 'writekey_sample')

        # check setter methods, explicitly
        _keen_provider._set_project_id('supguyz2')
        self.assertEqual(_keen_provider.project_id, 'supguyz2')
        self.assertEqual(_keen_provider.config['project_id'], 'supguyz2')

        _keen_provider._set_read_key('readkey_sample2')
        self.assertEqual(_keen_provider.read_key, 'readkey_sample2')
        self.assertEqual(_keen_provider.config['read_key'], 'readkey_sample2')

        _keen_provider._set_write_key('writekey_sample2')
        self.assertEqual(_keen_provider.write_key, 'writekey_sample2')
        self.assertEqual(_keen_provider.config['write_key'], 'writekey_sample2')

        # try to set an invalid config item
        with self.assertRaises(ValueError):
            _keen_provider._set_config('invalid_config', True)


## KeenUploaderTests - tests the :py:class:`KeenUploader`.
class KeenUploaderTests(KeenTests):

    ''' Tests the ``Keen`` uploader class, which is
        exposed at :py:class:`mix.KeenUploader`. '''

    def test_keen_upload(self):

        ''' Test the :py:meth:`upload` method on the uploader
            for ``Keen``, exposed at :py:class:`mix.KeenUploader`. '''

        _keen_provider = self._valid_keen_provider(instance=True)

        # check uploader cached access
        uploader = _keen_provider.uploader
        c_uploader = _keen_provider.uploader
        self.assertEqual(uploader, c_uploader)

        # make sure events are empty
        _keen_provider.uploader.events_by_kind = {}

        # make some sample events
        event_1 = 'sample_kind', {'blab': True}, None
        event_2 = 'sample_kind', {'bleebs': True}, mix.timezone('UTC').localize(datetime.datetime.fromtimestamp(random.randint(1377605000, 1377895000)))
        event_3 = 'sample_kind2', {'blops': True}, mix.timezone('UTC').localize(datetime.datetime.now())

        for kind, event, timestamp in (event_1, event_2, event_3):
            _keen_provider.uploader.upload(kind, json.dumps(event), timestamp if timestamp is not None else None)

        # check event kind existence
        self.assertTrue('sample_kind' in _keen_provider.uploader.events_by_kind)
        self.assertTrue('sample_kind2' in _keen_provider.uploader.events_by_kind)

        # check event kind length
        self.assertEqual(len(_keen_provider.uploader.events_by_kind['sample_kind']), 2)
        self.assertEqual(len(_keen_provider.uploader.events_by_kind['sample_kind2']), 1)

    def test_keen_upload_commit(self):

        ''' Test the :py:meth:`commit` method on the uploader
            for ``Keen``, exposed at :py:class:`mix.KeenUploader`. '''

        _keen_provider = self._valid_keen_provider(instance=True)

        # make sure events are empty
        _keen_provider.uploader.events_by_kind = {}

        # make some sample events
        event_1 = 'sample_kind', {'blab': True}, None
        event_2 = 'sample_kind', {'bleebs': True}, mix.timezone('UTC').localize(datetime.datetime.fromtimestamp(random.randint(1377605000, 1377895000)))
        event_3 = 'sample_kind2', {'blops': True}, mix.timezone('UTC').localize(datetime.datetime.now())

        for kind, event, timestamp in (event_1, event_2, event_3):
            _keen_provider.uploader.upload(kind, json.dumps(event), timestamp if timestamp is not None else None)

        # dispatch `commit`
        count = _keen_provider.uploader.commit()

        # make sure count matches
        self.assertEqual(count, 3)
        self.assertIsInstance(_keen_provider.lib.events_cache, dict)
        self.assertTrue('sample_kind' in _keen_provider.lib.events_cache)
        self.assertTrue('sample_kind2' in _keen_provider.lib.events_cache)
        self.assertEqual(len(_keen_provider.lib.events_cache['sample_kind']), 2)
        self.assertEqual(len(_keen_provider.lib.events_cache['sample_kind2']), 1)


## KeenDownloaderTests - tests the :py:class:`KeenDownloader`.
class KeenDownloaderTests(KeenTests):

    ''' Tests the ``Keen`` downloader class, which is
        exposed at :py;class:`mix.KeenDownloader`. '''

    def test_keen_download(self):

        ''' Test the :py:meth:`upload` method on the downloader
            for ``Keen``, exposed at :py:class:`mix.KeenUploader`. '''

        _keen_provider = self._valid_keen_provider(instance=True)

        # check downloader cached access
        downloader = _keen_provider.downloader
        c_downloader = _keen_provider.downloader
        self.assertEqual(downloader, c_downloader)

        # manufacture start and end datetimes
        start, end = (
            datetime.datetime(year=2013, month=8, day=20, hour=12, minute=0, second=0),
            datetime.datetime(year=2013, month=8, day=30, hour=12, minute=0, second=0)
        )

        # try to download events via generator
        events = []
        for kind, deserialized, timestamp in _keen_provider.downloader.download(start, end, ['sample_kind']):

            # check types
            self.assertIsInstance(kind, basestring)
            self.assertIsInstance(deserialized, dict)
            self.assertIsInstance(timestamp, datetime.datetime)

            # check values
            self.assertEqual(kind, 'sample_kind')

            # add to events
            events.append((kind, deserialized, timestamp))

        # check length of yielded events
        self.assertEqual(len(events), 2)



# -*- coding: utf-8 -*-

'''

    importer: mixpanel tests
    ~~~~~~~~~~~~~~~~~~~~~~~~

    this module tests ``Mixpanel``-specific provider functionality.

    :author: Sam Gammon <sam@keen.io>
    :license: This software follows the MIT (OSI-approved)
              license for open source software. A truncated
              version is included here; for full licensing
              details, see ``LICENSE.md`` in the root directory
              of the project.

              Copyright (c) 2013, Keen IO

              The above copyright notice and this permission notice shall be included in
              all copies or substantial portions of the Software.

              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
              IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
              FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
              AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
              LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
              OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
              THE SOFTWARE.

'''


# stdlib
import time
import json
import base64
import random
import unittest
import datetime

# importer
from importer import mix

# mixpanel constants
from importer.mix import _MIXPANEL_API
from importer.mix import _MIXPANEL_EXPORT

# main mixpanel-related classes
from importer.mix import Mixpanel
from importer.mix import MixpanelUploader
from importer.mix import MixpanelDownloader


## MixpanelTests - tests general integration with Mixpanel.
class MixpanelTests(unittest.TestCase):

    ''' Tests the ``Mixpanel`` provider class, which is
    	exposed at :py:class:`mix.Mixpanel`. '''

    random_ts = lambda self: datetime.datetime.fromtimestamp(random.randint(1377605000, 1377895000))

    def _valid_mixpanel_provider(self, instance=False, params=False):  # pragma: nocover

        ''' Tests the ``Mixpanel`` provider class, which is
            exposed at :py:class:`mix.Mixpanel`. '''

        ## `_request` shim
        def _request(self, methods, params={}):

            ''' Shim for submitting API requests to
                ``Mixpanel``. Returns ``self`` for
                chainability/introspection, unless
                the ``methods`` list contains 'export',
                in which case a fake export is mocked
                and returned. '''

            self.mock_request = (methods, params)
            if 'export' in methods:
                return "\n".join([
                    json.dumps({'event':'sample_kind','properties':{
                        'time': int(time.mktime(self.random_ts().timetuple())),
                        'blabs': True
                    }}),
                    json.dumps({'event':'sample_kind','properties':{
                        'time': int(time.mktime(self.random_ts().timetuple())),
                        'bleebs': True
                    }}),
                    ''  # throw in an extra line to try and juke the downloader
                ])  # return sample events

            return self  # otherwise return `self`

        ## `_library` shim
        def _library(self, api_key, api_secret):

            ''' Shim for constructing the API layer
                for ``Mixpanel``. Returns ``self`` for
                chainability/introspection. '''

            self.api_key = api_key
            self.api_secret = api_secret
            return self

        _bus_, _name_, _library_, _adapters_, _config_ = (
            object(),  # bus mock
            'MIXPANEL',  # provider name
            type('MixpanelMock', (object,), {
                'request': _request,  # request shim
                'api_key': None,  # fake API key
                'api_secret': None,  # fake API Secret
                'mock_request': None,  # fake request
                'Mixpanel': _library,  # fake constructor
                'random_ts': self.random_ts  # random timestamp lambda
            })(),  # library mock
            (MixpanelUploader, MixpanelDownloader),
            {
                'api_key': '_api_key_',  # mock mixpanel API key
                'api_secret': '_api_secret_'  # mock mixpanel API secret
            }
        )

        if instance:
            m = Mixpanel(_bus_, _name_, _library_, _adapters_, **_config_)
            if params:
                return (_bus_, _name_, _library_, _adapter_, _config_), m
            return m
        if params:
            return (_bus_, _name_, _library_, _adapters_, _config_), Mixpanel
        return Mixpanel

    def test_mixpanel_construct(self):

        ''' Test the :py:class:`Mixpanel` class, which
            implements the :py:class:`Provider`
            interface. '''

        _mixpanel_args, MixpanelProvider = self._valid_mixpanel_provider(instance=False, params=True)
        _bus, _name, _library, _adapters, _config = _mixpanel_args

        # try to construct without adapters
        with self.assertRaises(TypeError):
            MixpanelProvider(_bus, _name, _library, tuple(), **_config)

        # try to construct with incomplete adapters
        with self.assertRaises(TypeError):
            MixpanelProvider(_bus, _name, _library, (MixpanelUploader,), **_config)

        # try to construct in a valid way
        _valid_mixpanel = MixpanelProvider(_bus, _name, _library, _adapters, **_config)

        # test attached parameters
        self.assertNotEqual(_valid_mixpanel.timezone, None)

        # test attached API key and API secret
        self.assertEqual(_valid_mixpanel.lib.api_key, '_api_key_')
        self.assertEqual(_valid_mixpanel.lib.api_secret, '_api_secret_')

    def test_mixpanel_config(self):

        ''' Test :py:attr:`Mixpanel.config`, which exposes
            ``Mixpanel``-related provider config listed in
            *importer's* ``config.json``. '''

        _mixpanel_provider = self._valid_mixpanel_provider(instance=True)

        # check values from config
        self.assertIsInstance(_mixpanel_provider.config, dict)
        self.assertEqual(_mixpanel_provider.config.get('api_key'), '_api_key_')
        self.assertEqual(_mixpanel_provider.config.get('api_secret'), '_api_secret_')

    def test_mixpanel_toggle_export(self):

        ''' Test :py:attr:`Mixpanel` with Python's ``with``
            protocol, applied in this case to seamlessly
            switch the API endpoint used by ``Mixpanel``,
            which must be done when exporting versus
            importing events. '''

        _mixpanel_provider = self._valid_mixpanel_provider(instance=True)

        # should start at regular API enpdoint
        self.assertEqual(_mixpanel_provider.client.ENDPOINT, _MIXPANEL_API)

        # with blocks should properly re-raise exceptions
        with self.assertRaises(RuntimeError):

            with _mixpanel_provider:

                # should now be the export API enpdoint
                self.assertEqual(_mixpanel_provider.client.ENDPOINT, _MIXPANEL_EXPORT)
                raise RuntimeError('this should bubble up')

        self.assertEqual(_mixpanel_provider.client.ENDPOINT, _MIXPANEL_API)


## MixpanelUploaderTests - tests the :py:class:`MixpanelUploader`.
class MixpanelUploaderTests(MixpanelTests):

    ''' Tests the ``Mixpanel`` uploader class, which is
    	exposed at :py:class:`mix.MixpanelUploader`. '''

    def test_mixpanel_upload(self):

        ''' Test the :py:meth:`upload` method on the uploader
            for ``Mixpanel``, exposed at :py:class:`mix.MixpanelUploader`. '''

        _mixpanel_provider = self._valid_mixpanel_provider(instance=True)

        # check uploader cached access
        uploader = _mixpanel_provider.uploader
        c_uploader = _mixpanel_provider.uploader
        self.assertEqual(uploader, c_uploader)

        # make some sample events
        event_1 = 'sample_kind', {'blab': True}, None
        event_2 = 'sample_kind', {'bleebs': True}, datetime.datetime.fromtimestamp(random.randint(1377605000, 1377895000))
        event_3 = 'sample_kind2', {'blops': True}, datetime.datetime.now()

        for kind, event, timestamp in (event_1, event_2, event_3):
            _mixpanel_provider.uploader.upload(kind, json.dumps(event), mix._TO_MIXPANEL_DATE(timestamp) if timestamp else None)

        # check event kind existence
        self.assertTrue('sample_kind' in _mixpanel_provider.uploader.events_by_kind)
        self.assertTrue('sample_kind2' in _mixpanel_provider.uploader.events_by_kind)

        # check event kind length
        self.assertEqual(len(_mixpanel_provider.uploader.events_by_kind['sample_kind']), 2)
        self.assertEqual(len(_mixpanel_provider.uploader.events_by_kind['sample_kind2']), 1)

        return _mixpanel_provider

    def test_mixpanel_upload_commit(self):

        ''' Test the :py:meth:`commit` method on the uploader
            for ``Mixpanel``, exposed at :py:class:`mix.MixpanelUploader`. '''

        # bad form but why not: grab previous mixpanel provider
        _mixpanel_provider = self._valid_mixpanel_provider(instance=True)

        # make sure events are still there
        self.assertEqual(len(_mixpanel_provider.uploader.events_by_kind.keys()), 2)
        self.assertEqual(len(_mixpanel_provider.uploader.events_by_kind['sample_kind']), 2)
        self.assertEqual(len(_mixpanel_provider.uploader.events_by_kind['sample_kind2']), 1)

        # commit, see what happens
        count = _mixpanel_provider.uploader.commit()
        self.assertEqual(count, 3)

        # make sure there's a request
        self.assertNotEqual(_mixpanel_provider.client.mock_request, None)


## MixpanelDownloaderTests - tests the :py:class:`MixpanelDownloader`.
class MixpanelDownloaderTests(MixpanelTests):

    ''' Tests the ``Mixpanel`` downloader class, which is
    	exposed at :py:class:`mix.MixpanelDownloader`. '''

    def test_mixpanel_download(self):

        ''' Test the :py:meth:`upload` method on the downloader
            for ``Mixpanel``, exposed at :py:class:`mix.MixpanelUploader`. '''

        _mixpanel_provider = self._valid_mixpanel_provider(instance=True)

        # check downloader cached access
        downloader = _mixpanel_provider.downloader
        c_downloader = _mixpanel_provider.downloader
        self.assertEqual(downloader, c_downloader)

        # manufacture start and end datetimes
        start, end = (
            datetime.datetime(year=2013, month=8, day=20, hour=12, minute=0, second=0),
            datetime.datetime(year=2013, month=8, day=30, hour=12, minute=0, second=0)
        )

        # try to download events via generator
        events = []
        for kind, deserialized, timestamp in _mixpanel_provider.downloader.download(start, end, ['sample_kind']):

            # check types
            self.assertIsInstance(kind, basestring)
            self.assertIsInstance(deserialized, dict)
            self.assertIsInstance(timestamp, datetime.datetime)

            # check values
            self.assertEqual(kind, 'sample_kind')

            # add to events
            events.append((kind, deserialized, timestamp))

        # check length of yielded events
        self.assertEqual(len(events), 2)



# -*- coding: utf-8 -*-

'''

    importer: tests
    ~~~~~~~~~~~~~~~

    this module provides tests for the main ``importer``
    business logic module, provided at :py:mod:`importer.mix`.

    :author: Sam Gammon <sam@keen.io>
    :license: This software follows the MIT (OSI-approved)
              license for open source software. A truncated
              version is included here; for full licensing
              details, see ``LICENSE.md`` in the root directory
              of the project.

              Copyright (c) 2013, Keen IO

              The above copyright notice and this permission notice shall be included in
              all copies or substantial portions of the Software.

              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
              IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
              FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
              AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
              LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
              OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
              THE SOFTWARE.

'''


# stdlib
import sys
import logging
import unittest
import datetime

try:
    import cStringIO as StringIO
except ImportError:
    import StringIO

# tests
from importer_tests import test_keen
from importer_tests import test_mixpanel

# importer
from importer import tz
from importer import mix
from importer import __main__ as main

# main classes
from importer.mix import Importer
from importer.mix import Adapter
from importer.mix import Provider


## Globals
_MINIMAL_OUTPUT = '[importer]: Beginning event data download...\n[importer]: Beginning event data upload...\n[importer]: Committing events...\n[importer]: Upload complete.\n[importer]: Done, exiting `Importer`.\n'


## UtilTests - tests small utilities and lambdas provided by :py:mod:`mix`.
class UtilTests(unittest.TestCase):

    ''' Tests utility classes, lambdas, globals
        and functions. Random small stuff for
        format conversion and plumbing like that. '''

    def test_exports(self):

        ''' Tests that lambdas and constants bound
            at :py:mod:`importer.mix`'s top level
            exist and are accessible as attributes. '''

        for export in frozenset((

            # CSV-related globals
            '_CSV_FILE',  # file to output CSV buffer to
            '_CSV_PARAMS',  # config for :py:mod:`csv`

            # Mixpanel-related lambdas
            '_MIXPANEL_DATE_FMT',  # mixpanel's date format (YYYY-MM-DD)
            '_MIXPANEL_API',  # mixpanel's regular API endpoint
            '_MIXPANEL_EXPORT',  # mixpanel's export API endpoint

            # Keen-related lambdas
            '_KEEN_DATE_FMT',  # regular date format for keen (without TZ, for ``strptime``)
            '_KEEN_DATE_FMT_ISO',  # ISO8601 date format for keen (with TZ, for ``strftime``)

            '_PROVIDER',  # small lambda that resolves providers by name (for instance, _PROVIDER('keen'))
            '_PROVIDER_IMPL',  # small lambda that resolves a provider class by name, instead of a sentinel

            '_PRETTY_DATE',  # small lambda that formats a date in a pretty human-readable way
            '_TO_KEEN_DATE',  # formats a python ``datetime`` as a Keen-style ISO8601 timestamp
            '_FROM_KEEN_DATE',  # parses a Keen-style ISO8601 timestamp into a python ``datetime``
            '_TO_MIXPANEL_DATE',  # formats a python ``datetime`` as a Mixpanel-style timestamp (YYYY-MM-DD)
            '_FROM_MIXPANEL_DATE')):  # parses a Mixpanel-style timestamp into a python ``datetime``

            self.assertTrue(hasattr(mix, export))

    def test_mixpanel_constants(self):

        ''' Thorougly tests Mixpanel-related constants
            bound at :py:mod:`importer.mix`'s top level. '''

        ## == _MIXPANEL_DATE_FMT == ##

        # simple format test
        dt_now = datetime.datetime.now()
        dt_converted = datetime.datetime.strptime(dt_now.strftime(mix._MIXPANEL_DATE_FMT), mix._MIXPANEL_DATE_FMT)

        # drop resolution from `dt_now` and compare
        dt_now = datetime.datetime.combine(dt_now.date(), datetime.time(hour=0, minute=0))
        self.assertEqual(dt_now, dt_converted)  # check equality

        # try to unpack strings
        sample_valid, sample_invalid = '2013-01-05', '05-01-2013'

        # check valid date
        valid_dt = datetime.datetime.strptime(sample_valid, mix._MIXPANEL_DATE_FMT)
        self.assertEqual(valid_dt.year, 2013)
        self.assertEqual(valid_dt.month, 01)
        self.assertEqual(valid_dt.day, 5)

        # make sure we have our exceptions right
        with self.assertRaises(ValueError):
            invalid_dt = datetime.datetime.strptime(sample_invalid, mix._MIXPANEL_DATE_FMT)

        ## == _MIXPANEL_API / _MIXPANEL_EXPORT == ##

        # test regular endpoint
        self.assertIsInstance(mix._MIXPANEL_API, basestring)
        self.assertTrue('mixpanel.com' in mix._MIXPANEL_API)

        # test export endpoint
        self.assertIsInstance(mix._MIXPANEL_EXPORT, basestring)
        self.assertTrue('data.mixpanel.com' in mix._MIXPANEL_EXPORT)

    def test_keen_constants(self):
    
        ''' Thoroughly tests Keen-related constants
            bound at :py:mod:`importer.mix`'s top level. '''
    
        # sample DT
        sample = mix.timezone('UTC').localize(datetime.datetime(year=2013, month=8, day=24, hour=12, minute=30, second=01))

        # test extraction of a keen date
        sample_keen = mix._TO_KEEN_DATE(sample)

        # test formatting as a keen date
        interpreted_sample = mix._FROM_KEEN_DATE(sample_keen)

        # compare
        self.assertEqual(sample, interpreted_sample)

    def test_totimestamp(self):

        ''' Tests the top-level :py:func:`_TO_TIMESTAMP`
            lambda, which resolves an integer timestamp,
            given a Python :py:class:`datetime.datetime`. '''

        # setup known datetime and timestamp
        ts, dt = 1377372600, datetime.datetime(year=2013, month=8, day=24, hour=12, minute=30, second=0)
        self.assertEqual(ts, mix._TO_TIMESTAMP(dt))

    def test_provider_lambda(self):
    
        ''' Tests the top-level :py:func:`_PROVIDER`
            lambda, which resolves a provider by name. '''
    
        # make sure we can resolve keen
        self.assertEqual(mix.Providers.KEEN, 'KEEN')
        self.assertEqual(mix._PROVIDER('keen'), 'KEEN')
        self.assertEqual(mix._PROVIDER(mix.Providers.KEEN), 'KEEN')
        self.assertEqual(mix._PROVIDER_IMPL('KEEN'), mix.Keen)
        self.assertEqual(mix._PROVIDER_IMPL(mix.Providers.KEEN), mix.Keen)
        self.assertEqual(mix._PROVIDER_IMPL(mix._PROVIDER('keen')), mix.Keen)

        # make sure we can resolve mixpanel
        self.assertEqual(mix.Providers.MIXPANEL, 'MIXPANEL')
        self.assertEqual(mix._PROVIDER('mixpanel'), 'MIXPANEL')
        self.assertEqual(mix._PROVIDER(mix.Providers.MIXPANEL), 'MIXPANEL')
        self.assertEqual(mix._PROVIDER_IMPL('MIXPANEL'), mix.Mixpanel)
        self.assertEqual(mix._PROVIDER_IMPL(mix.Providers.MIXPANEL), mix.Mixpanel)
        self.assertEqual(mix._PROVIDER_IMPL(mix._PROVIDER('mixpanel')), mix.Mixpanel)

    def test_pretty_date(self):
    
        ''' Tests the top-level :py:func:`_PRETTY_DATE`
            lambda, which formats a date in a nice way. '''
    
        # try to format a known date in a pretty way
        x = datetime.datetime(year=2013, month=8, day=24, hour=12, minute=30)
        self.assertEqual(mix._PRETTY_DATE(x), 'Sat Aug 24, 2013')

    def test_mixpanel_lambdas(self):
    
        ''' Test Mixpanel-related lambdas, which are
            exposed at :py:func:`_TO_MIXPANEL_DATE`
            and :py:func:`_FROM_MIXPANEL_DATE`. '''
    
        # make a known date
        sample_mixpanel_dt = datetime.datetime(year=2013, month=8, day=24)

        # try using the mixpanel format
        mixpanel_fmt = sample_mixpanel_dt.strftime(mix._MIXPANEL_DATE_FMT)
        self.assertEqual(mixpanel_fmt, '2013-08-24')

        # try using the lambdas
        mixpanel_date = mix._FROM_MIXPANEL_DATE(mixpanel_fmt)
        self.assertTrue(sample_mixpanel_dt, mixpanel_date)

        # try a full deserialize->serialize
        self.assertTrue(sample_mixpanel_dt, mix._FROM_MIXPANEL_DATE(mix._TO_MIXPANEL_DATE(sample_mixpanel_dt)))


## ProviderTests - tests functionality related to abstract providers.
class ProviderTests(unittest.TestCase):

    ''' Tests the abstract provider class, which
        is exposed at :py:class:`mix.Provider`. '''

    @classmethod
    def _valid_provider(cls, instance=False, params=False, klass=True):  # pragma: nocover

        ''' Small utility method to generate a valid
            :py:class:`mix.Provider` implementation
            on-the-fly. '''

        # can only return an instance or params as 2nd pos value
        if instance and params: raise TypeError('you are doing it wrong')

        ## == Valid Provider == ##

        class ValidProvider(mix.Provider):

            ''' Sample (valid) provider implementation,
                tests ABC enforcement. '''

            def initialize(self):

                ''' Test initialization method, which is
                    enforced by the :py:class:`mix.Provider`
                    spec. '''

                return

        class ValidUploader(mix.Uploader):

            ''' Sample (valid) uploader implementation,
                for testing property access. '''

            def commit(self, *args, **kwargs):

                ''' Commit hook. Must be defined to
                    provide an uploader. '''

            def upload(self, *args, **kwargs):

                ''' Upload hook. Must be defined to
                    provide an uploader. '''

        class ValidDownloader(mix.Downloader):

            ''' Sample (valid) downloader implementation,
                for testing property access. '''

            def download(self):

                ''' Download hook. Must be defined to
                    provide a downloader. '''

        # provider obj
        _importer_, _name_, _library_, _adapters_, _config_ = (
            object(),
            'VALID',
            object(),
            (ValidUploader, ValidDownloader),
            {'sample_key': True}
        )

        if instance:
            if klass:
                return ValidProvider, ValidProvider(_importer_, _name_, _library_, _adapters_, **_config_)
            return ValidProvider(_importer_, _name_, _library_, _adapters_, **_config_)
        if params:
            if klass:
                return ValidProvider, (_importer_, _name_, _library_, _adapters_, _config_)
            return (_importer_, _name_, _library_, _adapters_, _config_)
        return ValidProvider

    def test_provider_abc(self):

        ''' Test Abstract Base Class semantics, against
            an invalid and valid :py:class:`mix.Provider`
            interface. '''

        ValidProvider = self._valid_provider(instance=False)

        # tests
        self.assertTrue(ValidProvider)  # is it there?
        self.assertIsInstance(ValidProvider, type)  # not a singleton?
        self.assertTrue(issubclass(ValidProvider, mix.Provider))  # inheritance

        ## == Invalid Provider == ##

        with self.assertRaises(TypeError):

            class InvalidProvider(mix.Provider):

                ''' Sample (invalid) provider implementation,
                    tests ABC enforcement. '''

                pass  # should have an ``initialize`` method

            invalid = InvalidProvider()  # should fail

    def test_provider_map(self):

        ''' Tests the top-level provider map, exported
            from :py:mod:`mix` as :py:class:`mix.Providers`. '''

        self.assertTrue(hasattr(mix, 'Providers'))  # class existence
        self.assertTrue(hasattr(mix.Providers, 'KEEN'))  # `KEEN` mapping
        self.assertTrue(hasattr(mix.Providers, 'MIXPANEL'))  # `MIXPANEL` mapping

    def test_provider_repr(self):

        ''' Tests :py:class:`Provider`'s ability to
            generate a nice string representation of
            itself. '''

        # test __repr__
        ValidProvider, _valid_provider = self._valid_provider(instance=True)
        self.assertEqual(_valid_provider.__repr__(), 'Provider(VALID)')

    def test_provider_workers(self):

        ''' Tests :py:class:`Provider`'s proxied lazy
            properties for accessing attached/bound
            :py:class:`Uploader` and :py:class:`Downloader`
            objects. '''

        ValidProvider, _params = self._valid_provider(params=True)
        _importer, _name, _library, _adapters, _config = _params

        # instantiate provider
        _provider = ValidProvider(*(_importer, _name, _library, _adapters), **_config)

        # up/downloader
        _up, _down = _adapters[0], _adapters[1]

        # test workers
        self.assertIsInstance(_provider.uploader, _up)
        self.assertIsInstance(_provider.uploader, _up)  # tests cached access
        self.assertIsInstance(_provider.uploader, mix.Uploader)
        self.assertIsInstance(_provider.downloader, _down)
        self.assertIsInstance(_provider.downloader, _down)  # tests cached access
        self.assertIsInstance(_provider.downloader, mix.Downloader)

    def test_provider_logging(self):

        ''' Tests the logging interface provided internally
            for :py:class:`Provider` implementors. '''

        ValidProvider, _valid_provider = self._valid_provider(instance=True)

        # make sure logging is there and looks like a logging pipe
        self.assertTrue(hasattr(_valid_provider.logging, 'debug'))
        self.assertTrue(hasattr(_valid_provider.logging, 'info'))
        self.assertTrue(hasattr(_valid_provider.logging, 'warning'))
        self.assertTrue(hasattr(_valid_provider.logging, 'error'))
        self.assertTrue(hasattr(_valid_provider.logging, 'critical'))


## AdapterTests - tests functionality related to abstract adapters.
class AdapterTests(unittest.TestCase):

    ''' Tests the abstract adapter class, which
        is exposed at :py:class:`mix.Adapter`. '''

    @classmethod
    def _valid_adapter(cls, uploader=False, downloader=True, instance=False):  # pragma: nocover

        ''' Acquire a valid :py:class:`Uploader`,
            :py:class:`Downloader`, or both. Whether
            an instance or class is returned is
            switchable via the kwarg ``instance``. '''

        _result = []

        if uploader:
            class ValidUploader(mix.Uploader):

                ''' Sample (valid) uploader implementation,
                    for testing property access. '''

                def commit(self, *args, **kwargs):

                    ''' Commit hook. Must be defined to
                        provide an uploader. '''

                def upload(self, *args, **kwargs):

                    ''' Upload hook. Must be defined to
                        provide an uploader. '''

            if instance:
                _result.append(ValidUploader(ProviderTests._valid_provider(instance=True, klass=False)))
            else:
                _result.append(ValidUploader)

        if downloader:
            class ValidDownloader(mix.Downloader):

                ''' Sample (valid) downloader implementation,
                    for testing property access. '''

                def download(self):

                    ''' Download hook. Must be defined to
                        provide a downloader. '''

            if instance:
                _result.append(ValidDownloader(ProviderTests._valid_provider(instance=True, klass=False)))
            else:
                _result.append(ValidDownloader)

        if _result:
            if len(_result) > 1:
                return tuple(_result)
            return _result[0]

    def test_adapter_direct_extend(self):

        ''' Make sure that new adapter types can't
            be added - only :py:class:`Uploader`
            and :py:class:`Downloader` may extend
            :py:class:`Adapter`. '''
    
        with self.assertRaises(RuntimeError):
            # try to extend :py:class:`Adapter` in an invalid way
            class SomeRandomAdapter(mix.Adapter):
                pass

    def test_adapter_direct(self):
    
        ''' Try directly instantiating a valid
            adapter, and handing it a mock
            :py:class:`Provider`. '''
    
        # try to instantiate abstract adapters directly
        with self.assertRaises(TypeError):
            mix.Adapter()
        with self.assertRaises(TypeError):
            mix.Uploader()
        with self.assertRaises(TypeError):
            mix.Downloader()

        # try instantiating a valid uploader/etc without a provider (direct)
        with self.assertRaises(RuntimeError):
            mix.KeenDownloader()
        with self.assertRaises(RuntimeError):
            mix.KeenUploader()

    def test_adapter_logging(self):

        ''' Make sure that the internal logging
            interface provided to :py:class:`Adapter`
            implementors works correctly. '''

        _uploader, _downloader = self._valid_adapter(uploader=True, downloader=True, instance=True)

        # make sure logging is logging-like (u/l)
        self.assertTrue(hasattr(_uploader.logging, 'debug'))
        self.assertTrue(hasattr(_uploader.logging, 'info'))
        self.assertTrue(hasattr(_uploader.logging, 'warning'))
        self.assertTrue(hasattr(_uploader.logging, 'error'))
        self.assertTrue(hasattr(_uploader.logging, 'critical'))

        # make sure logging is logging-like (d/l)
        self.assertTrue(hasattr(_downloader.logging, 'debug'))
        self.assertTrue(hasattr(_downloader.logging, 'info'))
        self.assertTrue(hasattr(_downloader.logging, 'warning'))
        self.assertTrue(hasattr(_downloader.logging, 'error'))
        self.assertTrue(hasattr(_downloader.logging, 'critical'))

    def test_adapter_transform(self):

        ''' Make sure that :py:meth:`Adapter.transform`
            works correctly as a soft-abstract method. '''

        _uploader, _downloader = self._valid_adapter(uploader=True, downloader=True, instance=True)

        # make sure transform is there
        self.assertTrue(hasattr(_uploader, 'transform'))
        self.assertTrue(hasattr(_downloader, 'transform'))

        # make sure it returns ``NotImplemented``
        self.assertEqual(_uploader.transform({}), NotImplemented)
        self.assertEqual(_downloader.transform({}), NotImplemented)

        # make sure the abstract method raises ``NotImplementedError``
        with self.assertRaises(NotImplementedError):
            super(mix.Uploader, _uploader).transform({})

        with self.assertRaises(NotImplementedError):
            super(mix.Downloader, _downloader).transform({})


## UploaderTests - tests functionality specific to abstract uploaders.
class UploaderTests(AdapterTests):

    ''' Tests the abstract uploader class, which
        is exposed at :py:class:`mix.Uploader`. '''

    def test_uploader_abc(self):

        ''' Test Abstract Base Class semantics, against
            an invalid and valid :py:class:`mix.Uploader`
            interface. '''

        ## == Invalid Uploader == ##
        with self.assertRaises(TypeError):
            class InvalidUploader(mix.Uploader):
                pass

            InvalidUploader()

        ## == Valid Uploader == ##

        # valid instantiation: with provider
        self._valid_adapter(uploader=True, downloader=False, instance=False)(object())

    def test_uploader_direct(self):

        ''' Test direct use of an :py:class:`Uploader`
            class, which is prohibited. '''

        # invalid instantiation: no provider
        with self.assertRaises(RuntimeError):
            self._valid_adapter(uploader=True, downloader=False, instance=False)()

        # valid instantiation: with provider
        self._valid_adapter(uploader=True, downloader=False, instance=False)(object())

    def test_uploader_preprocess(self):

        ''' Test the :py:meth:`pre_process` method, which
            dispatches :py:meth:`transform` and hands post-
            tranform data back to the adapter for buffering.  '''

        # generate valid adapter to test transform
        ValidUploader = self._valid_adapter(uploader=True, downloader=False, instance=False)
        _valid_uploader = self._valid_adapter(uploader=True, downloader=False, instance=True)

        class ValidTransformUploader(ValidUploader, mix.Uploader):  # must mix-in Uploader to add to enums

            ''' Sample (valid) uploader implementation,
                for testing property access. '''

            def transform(self, data):
                
                ''' Testable, implemented :py:meth:`transform`
                    definition. Used to test non-null transforms. '''

                data['prop'] = not data['prop']  # make sure we can read + write
                data['new_prop'] = 'sup'  # make sure we can add stuff
                return data

        # sample pre-transform data
        data = {'prop': True, 'preserved_prop': 123}

        # try a null-op transform, make sure it defaults to ``NotImplemented``
        result = _valid_uploader.transform(dict(data.items()[:]))

        self.assertEqual(result, NotImplemented)  # should return, NOT raise
        self.assertEqual(data['prop'], True)  # make sure nothing changed
        self.assertEqual(data['preserved_prop'], 123)  # with either of our props

        # shouldn't get any new keys yet
        with self.assertRaises(KeyError):
            data['new_prop']

        # try a valid transform operation
        _transform_uploader = ValidTransformUploader(object())
        transformed_result = _transform_uploader.transform(dict(data.items()[:]))

        self.assertIsInstance(transformed_result, dict)
        self.assertEqual(transformed_result['prop'], (not data['prop']))
        self.assertEqual(transformed_result['new_prop'], 'sup')
        self.assertEqual(transformed_result['preserved_prop'], data['preserved_prop'])

        # make sure the pre-process wrapper works
        noop_transform = _valid_uploader.pre_process(dict(data.items()[:]))

        # noop-transformed data should remain the same
        self.assertEqual(noop_transform['prop'], data['prop'])
        self.assertEqual(noop_transform['preserved_prop'], data['preserved_prop'])

        with self.assertRaises(KeyError):
            noop_transform['new_prop']

        # try a valid pre_process operation
        valid_transform = _transform_uploader.pre_process(dict(data.items()[:]))

        # valid-transform data should come back transformed
        self.assertEqual(valid_transform['prop'], (not data['prop']))
        self.assertEqual(valid_transform['new_prop'], 'sup')
        self.assertEqual(valid_transform['preserved_prop'], data['preserved_prop'])


## DownloaderTests - tests functionality specific to abstract downloaders.
class DownloaderTests(AdapterTests):

    ''' Tests the abstract downloader class, which
        is exposed at :py:class:`mix.Downloader`. '''

    def test_downloader_abc(self):

        ''' Test Abstract Base Class semantics, against
            an invalid and valid :py:class:`mix.Downloader`
            interface. '''

        ## == Invalid Uploader == ##
        with self.assertRaises(TypeError):
            class InvalidDownloader(mix.Downloader):
                pass

            InvalidDownloader()

        ## == Valid Uploader == ##

        # valid instantiation: with provider
        self._valid_adapter(uploader=False, downloader=True, instance=False)(object())

    def test_downloader_direct(self):

        ''' Test direct use of an :py:class:`Uploader`
            class, which is prohibited. '''

        # invalid instantiation: no provider
        with self.assertRaises(RuntimeError):
            self._valid_adapter(uploader=False, downloader=True, instance=False)()

        # valid instantiation: with provider
        self._valid_adapter(uploader=False, downloader=True, instance=False)(object())

    def test_downloader_postprocess(self):

        ''' Test the :py:meth:`post_process` method, which
            dispatches :py:meth:`transform` and hands post-
            tranform data back to the adapter for buffering.  '''

        # generate valid adapter to test transform
        ValidDownloader = self._valid_adapter(uploader=False, downloader=True, instance=False)
        _valid_downloader = self._valid_adapter(uploader=False, downloader=True, instance=True)

        class ValidTransformDownloader(ValidDownloader, mix.Downloader):  # must mix-in Uploader to add to enums

            ''' Sample (valid) downloader implementation,
                for testing property access. '''

            def transform(self, data):
                
                ''' Testable, implemented :py:meth:`transform`
                    definition. Used to test non-null transforms. '''

                data['prop'] = not data['prop']  # make sure we can read + write
                data['new_prop'] = 'sup'  # make sure we can add stuff
                return data

        # sample pre-transform data
        data = {'prop': True, 'preserved_prop': 123}

        # try a null-op transform, make sure it defaults to ``NotImplemented``
        result = _valid_downloader.transform(dict(data.items()[:]))

        self.assertEqual(result, NotImplemented)  # should return, NOT raise
        self.assertEqual(data['prop'], True)  # make sure nothing changed
        self.assertEqual(data['preserved_prop'], 123)  # with either of our props

        # shouldn't get any new keys yet
        with self.assertRaises(KeyError):
            data['new_prop']

        # try a valid transform operation
        _transform_downloader = ValidTransformDownloader(object())
        transformed_result = _transform_downloader.transform(dict(data.items()[:]))

        self.assertIsInstance(transformed_result, dict)
        self.assertEqual(transformed_result['prop'], (not data['prop']))
        self.assertEqual(transformed_result['new_prop'], 'sup')
        self.assertEqual(transformed_result['preserved_prop'], data['preserved_prop'])

        # make sure the pre-process wrapper works
        noop_transform = _valid_downloader.post_process(dict(data.items()[:]))

        # noop-transformed data should remain the same
        self.assertEqual(noop_transform['prop'], data['prop'])
        self.assertEqual(noop_transform['preserved_prop'], data['preserved_prop'])

        with self.assertRaises(KeyError):
            noop_transform['new_prop']

        # try a valid pre_process operation
        valid_transform = _transform_downloader.post_process(dict(data.items()[:]))

        # valid-transform data should come back transformed
        self.assertEqual(valid_transform['prop'], (not data['prop']))
        self.assertEqual(valid_transform['new_prop'], 'sup')
        self.assertEqual(valid_transform['preserved_prop'], data['preserved_prop'])


## FlowTests - tests :py:class:`mix.Importer` data flow.
class FlowTests(test_keen.KeenTests, test_mixpanel.MixpanelTests):

    ''' Tests flow of data and order of method execution
        in :py:class:`Importer` proper. '''

    def _valid_importer(self, instance=False, params=False, _full=False):  # pragma: nocover

        ''' Generate a valid :py:class:`mix.Importer` instance,
            injecting mock parameters as needed. Optionally
            provide a full instance or just the parameters,
            in place of a valid ``Importer`` class. '''

        if not instance and not params:  # pragma: nocover
            return mix.Importer

        if not _full:  # pragma: nocover
            # generate params
            _keen_ = type('KeenMock', (object,), {
                'project_id': None,
                'read_key': None,
                'write_key': None
            })()

            _mixpanel_ = type('MixpanelMock', (object,), {
                'api_key': None,
                'secret_key': None
            })()

            if params and not instance:
                return (_keen_, _mixpanel_), mix.Importer
            return mix.Importer(_keen_, _mixpanel_)

        else:  # pragma: nocover

            # construct valid, mocked providers
            _keen_params, _keen = self._valid_keen_provider(False, True)
            _k_bus, _k_name, _k_library, _k_adapters, _k_config = _keen_params

            _mixpanel_params, _mixpanel = self._valid_mixpanel_provider(False, True)
            _m_bus, _m_name, _m_library, _m_adapters, _m_config = _mixpanel_params

            if params and not instance:
                return (_k_library, _m_library)
            return mix.Importer(_k_library, _m_library)

    def test_importer_construct(self):

        ''' Test direct construction of the
            :py:class:`Importer` object, which is
            the preferred way to dispatch ``importer``
            directly from Python. '''

        _importer_params, Importer = self._valid_importer(params=True)
        _keen, _mixpanel = _importer_params

        # make sure things are not initialized and proper defaults are applied
        self.assertEqual(Importer.cli, False)  # `CLI` mode should default to `False`
        self.assertEqual(Importer.yes, True)  # `yes` mode should default to `False`
        self.assertEqual(Importer.debug, False)  # `debug` mode should default to `False`
        self.assertEqual(Importer.quiet, False)  # `quiet` mode should default to `False`
        self.assertEqual(Importer.log_level, logging.INFO)  # `log_level` should default to `INFO`

        # grab lib ref to mixpanel and remove from visibility
        if 'mixpanel' in sys.modules:
            _mixpanel_mod = sys.modules['mixpanel']
            sys.modules['mixpanel'] = None
        else:
            _mixpanel_mod = None

        # grab lib ref to keen and remove from visibility
        if 'keen' in sys.modules:
            _keen_mod = sys.modules['keen']
            sys.modules['keen'] = None
        else:
            _keen_mod = None

        def _construct_importer(self, keen=None, mixpanel=None, strict=False, _do_check=True):

            ''' Constructor shim for ``Importer`` to help
                with construction testing. '''

            m = Importer(keen, mixpanel, strict=strict)

            if _do_check and m:
                if keen:
                    self.assertTrue('keen' in m)
                else:
                    self.assertFalse('keen' in m)
                if mixpanel:
                    self.assertTrue('mixpanel' in m)
                else:
                    self.assertFalse('mixpanel' in m)

            return m

        ### === STRICT MODE TESTS === ###

        # try strict with no libraries and no global context
        with self.assertRaises(ImportError):
            _construct_importer(self, None, None, True, False)

        # try strict with only keen and no global context
        with self.assertRaises(ImportError):
            _construct_importer(self, _keen, None, True, False)

        # try strict with only mixpanel and no global context
        with self.assertRaises(ImportError):
            _construct_importer(self, None, _mixpanel, True, False)

        # try strict with both libraries and no global context
        _construct_importer(self, _keen, _mixpanel, True, True)

        if _keen_mod:  # we have keen held globally
            
            sys.modules['keen'] = _keen_mod

            # try strict with no libraries and keen in the global context
            with self.assertRaises(ImportError):
                _construct_importer(self, None, None, True, False)

            # try strict with keen only and keen in the global context
            with self.assertRaises(ImportError):
                _construct_importer(self, _keen, None, True, False)

            # try strict with mixpanel only and keen in the global context
            m = _construct_importer(self, None, _mixpanel, True, False)
            self.assertTrue('keen' in m)
            self.assertTrue('mixpanel' in m)

            # try strict with keen and mixpanel and keen in the global context
            m = _construct_importer(self, _keen, _mixpanel, True, False)
            args, provider = m.provider('keen', False, True)
            library, provider, uploader, downloader = args  # extract args

            # perform tests to make sure locally-handed ``keen`` overrides the global copy
            self.assertTrue('keen' in m)
            self.assertTrue('mixpanel' in m)
            self.assertEqual(m.provider('keen', True).lib, _keen)

            # remove keen from global context again
            _keen_mod = sys.modules['keen']
            sys.modules['keen'] = None

        if _mixpanel_mod:

            sys.modules['mixpanel'] = _mixpanel_mod

            # try strict with no libraries and mixpanel in the global context
            with self.assertRaises(ImportError):
                _construct_importer(self, None, None, True, False)

            # try strict with mixpanel only and mixpanel in the global context
            with self.assertRaises(ImportError):
                _construct_importer(self, None, _mixpanel, True, False)

            # try strict with keen only and mixpanel in the global context
            m = _construct_importer(self, _keen, None, True, False)
            self.assertTrue('keen' in m)
            self.assertTrue('mixpanel' in m)

            # try strict with keen and mixpanel and mixpanel in the global context
            m = _construct_importer(self, _keen, _mixpanel, True, False)
            args, provider = m.provider('mixpanel', False, True)
            library, provider, uploader, downloader = args

            # perform tests to make sure locally-handed ``mixpanel`` overrides the global copy
            self.assertTrue('keen' in m)
            self.assertTrue('mixpanel' in m)
            self.assertEqual(library, _mixpanel)

            # remove mixpanel from global context again
            _mixpanel_mod = sys.modules['mixpanel']
            sys.modules['mixpanel'] = None

        if _keen_mod and _mixpanel_mod:

            sys.modules['mixpanel'], sys.modules['keen'] = _mixpanel_mod, _keen_mod

            # try strict with no libraries and both in the global context
            m = _construct_importer(self, None, None, True, False)
            
            # get keen provider stuff
            k_args, k_provider = m.provider('keen', False, True)
            k_lib, k_provider_d, k_up, k_down = k_args

            # get mixpanel provider stuff
            m_args, m_provider = m.provider('mixpanel', False, True)
            m_lib, m_provider_d, m_up, m_down = m_args

            self.assertTrue('keen' in m)
            self.assertTrue('mixpanel' in m)

            # completely clear context and check mount from direct `import`
            sys.modules['mixpanel'], sys.modules['keen'] = None, None
            del sys.modules['mixpanel']
            del sys.modules['keen']

            # try constructing completely from valid global context
            m = _construct_importer(self, None, None, True, False)

            # get keen provider stuff
            k_args, k_provider = m.provider('keen', False, True)
            k_lib, k_provider_d, k_up, k_down = k_args

            # get mixpanel provider stuff
            m_args, m_provider = m.provider('mixpanel', False, True)
            m_lib, m_provider_d, m_up, m_down = m_args

            self.assertTrue('keen' in m)
            self.assertTrue('mixpanel' in m)

    def test_importer_provider(self):

        ''' Make sure that ``Importer`` can hand back
            providers when asked. '''

        _importer = self._valid_importer(instance=True)

        # test regular provider class access
        _keen_klass, _mixpanel_klass = _importer.provider('keen', instance=False), _importer.provider('mixpanel', instance=False)

        # make sure we're handed back classes
        self.assertIsInstance(_keen_klass, type)
        self.assertIsInstance(_mixpanel_klass, type)

        # try to ask for a provider that doesn't exist
        with self.assertRaises(NotImplementedError):
            _importer.provider('blabble', instance=False)

    def test_importer_direct_execution(self):

        ''' Test :py:class:`Importer`'s direct dispatch
            interface, by manually constructing and moving
            through a data flow. Stub out the endpoints/etc
            to allow mock providers. '''

        _importer = self._valid_importer(instance=True, _full=True)

        # prepare args and try an execution flow
        _begin, _end, _args = (
            datetime.datetime(year=2013, month=8, day=20, hour=12, minute=0, second=0),
            datetime.datetime(year=2013, month=8, day=30, hour=12, minute=0, second=0),
            {
                'yes': True,  # skip prompts
                'quiet': False,  # don't silence output
                'verbose': False,  # don't increase output
                'kinds': ['sample_kind'],  # only download sample kind
                'reset_ts': False,  # don't reset timestamps
                'to': 'keen',  # mock a transfer to keen
                'from': 'mixpanel'  # mock a transfer from mixpanel
            }
        )

        # try a flow, because why not
        _importer.execute(_begin, _end, **_args)

    def test_importer_cli_execution(self):

        ''' Simulate a call to :py:class:`Importer` from the
            command line. Make sure that arguments are
            parsed right and the call goes to the right
            ``Importer`` entrypoint (``__call__``). '''

        _importer = self._valid_importer(instance=True, _full=True)

        # prepare args and try an execution flow
        _cliflags = {
            'begin': mix._TO_MIXPANEL_DATE(datetime.datetime(year=2013, month=8, day=20, hour=12, minute=0, second=0)),
            'end': mix._TO_MIXPANEL_DATE(datetime.datetime(year=2013, month=8, day=30, hour=12, minute=0, second=0)),
            'yes': True,  # skip prompts
            'quiet': False,  # don't silence output
            'verbose': False,  # don't increase output
            'kinds': ['sample_kind'],  # only download sample kind
            'reset_ts': False,  # don't reset timestamps
            'to': 'keen',  # mock a transfer to keen
            'from': 'mixpanel'  # mock a transfer from mixpanel
        }

        # try a flow, because why not
        _importer(**_cliflags)


## CLITests - tests :py:class:`mix.Importer` CLI usage.
class CLITests(FlowTests):

    ''' Tests :py:class:`Importer` functionality related to
        command line dispatch and argument parsing. '''

    def _valid_importer_cli(self):

        ''' Simulate a call to :py:class:`Importer` from the
            command line. Make sure that arguments are
            parsed right and the call goes to the right
            ``Importer`` entrypoint (``__call__``). '''

        _importer = self._valid_importer(instance=True, _full=True)

        # prepare args and try an execution flow
        _cliflags = {
            'begin': mix._TO_MIXPANEL_DATE(datetime.datetime(year=2013, month=8, day=20, hour=12, minute=0, second=0)),
            'end': mix._TO_MIXPANEL_DATE(datetime.datetime(year=2013, month=8, day=30, hour=12, minute=0, second=0)),
            'yes': True,  # skip prompts
            'quiet': False,  # don't silence output
            'verbose': False,  # don't increase output
            'kinds': ['sample_kind'],  # only download sample kind
            'reset_ts': False,  # don't reset timestamps
            'to': 'keen',  # mock a transfer to keen
            'from': 'mixpanel'  # mock a transfer from mixpanel
        }

        return _cliflags, _importer

    def test_importer_quiet(self):

        ''' Test the ``-q``/``--quiet`` flags to ``Importer``. '''

        flags, mix = self._valid_importer_cli()

        # run with quiet flag
        flags['quiet'] = True

        # switch stdout with a shim
        original_stdout, sys.stdout = sys.stdout, StringIO.StringIO()

        mix(**flags)

        # switch stdout again
        shim_stdout, sys.stdout = sys.stdout, original_stdout

        self.assertEqual(len(shim_stdout.getvalue()), 188)
        self.assertEqual(shim_stdout.getvalue(), _MINIMAL_OUTPUT)

    def test_importer_verbose(self):

        ''' Test the ``-v``/``--verbose`` flags to ``Importer``. '''

        flags, mix = self._valid_importer_cli()

        # run with quiet flag
        flags['quiet'] = False
        flags['verbose'] = True

        # switch stdout with a shim
        original_stdout, original_stderr, sys.stdout, sys.stderr = sys.stdout, sys.stderr, StringIO.StringIO(), StringIO.StringIO()

        mix(**flags)

        # switch stdout again
        shim_stdout, shim_stderr, sys.stdout, sys.stderr = sys.stdout, sys.stderr, original_stdout, original_stderr

        self.assertTrue(shim_stdout.getvalue() != _MINIMAL_OUTPUT)
        self.assertTrue(len(shim_stdout.getvalue() + shim_stderr.getvalue()) > 182)



# -*- coding: utf-8 -*-

'''

    importer: testrunner
    ~~~~~~~~~~~~~~~~~~~~

    this file collects and runs ``importer``'s testsuite.

    :author: Sam Gammon <sam@keen.io>
    :license: This software follows the MIT (OSI-approved)
              license for open source software. A truncated
              version is included here; for full licensing
              details, see ``LICENSE.md`` in the root directory
              of the project.

              Copyright (c) 2013, Keen IO

              The above copyright notice and this permission notice shall be included in
              all copies or substantial portions of the Software.

              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
              IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
              FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
              AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
              LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
              OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
              THE SOFTWARE.

'''

# pragma: nocover

# stdlib
import os
import sys
import unittest


## Globals

_TEST_PATHS = [
    'test_tz',  # timezone-related tests
    'test_mix',  # importer-related tests (and abstract stuff)
    'test_keen',  # keen-specific adapter/provider tests
    'test_mixpanel'  # mixpanel-specific adapter/provider tests
]


## Testsuite

## `fix_path`: add local paths to sys.path
def fix_path():

    ''' Add ``.`` and ``dir/lib`` to ``sys.path``. '''

    for path in ((os.path.dirname(__file__)),
                 (os.path.dirname(os.path.dirname(__file__))),
                 (os.path.join(os.path.dirname(os.path.dirname(__file__)), 'lib'))):

        sys.path.insert(0, path)


## `load_module` - Load a single testsuite module.
def load_module(path):

    ''' __main__ testing entrypoint for `apptools.model`. '''

    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromName(path))
    return suite


## `load` - Gather AppTools testsuites.
def load(paths=None):

    ''' __main__ entrypoint '''

    mixtests = unittest.TestSuite()

    if paths is None:
        paths = _TEST_PATHS[:]

    for path in paths:
        mixtests.addTest(load_module(path))

    return mixtests


## `run` - Run a suite of tests loaded via `load`.
def run(suite=None):

    ''' Optionally allow switching between XML or text output, if supported. '''

    fix_path()
    if suite is None:
        suite = load()

    # resolve text runner
    if hasattr(unittest, 'TestRunner'): runner = unittest.TestRunner
    else: runner = unittest.TextTestRunner

    if len(sys.argv) > 1:
        args = sys.argv[1:]  # slice off invocation

        if len(args) == 2:  # <format>, <output location>
            format, output = tuple(args)

            if format.lower().strip() == 'xml':
                try:
                    import xmlrunner
                except ImportError:
                    print "ERROR! XML testrunner (py module `xmlrunner`) could not be imported. Please run in text-only mode."
                    exit(1)
                xmlrunner.XMLTestRunner(output=output).run(suite)

            elif format.lower().strip() == 'text':
                return runner(verbosity=5, output=output).run(suite)
        else:
            return runner(verbosity=5, output=output).run(suite)  # text mode with

    return runner(verbosity=5).run(suite)



# -*- coding: utf-8 -*-

'''
    importer
    ~~~~~~~~

    a handy utility for transferring event data
    between ``keen.io`` and ``mixpanel``.

    usage: ``python mixpanel -h``
    (from outside the project directory)

    :author: Sam Gammon <sam@keen.io>
    :license: This software follows the MIT (OSI-approved)
              license for open source software. A truncated
              version is included here; for full licensing
              details, see ``LICENSE.md`` in the root directory
              of the project.

              Copyright (c) 2013, Keen IO

              The above copyright notice and this permission notice shall be included in
              all copies or substantial portions of the Software.

              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
              IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
              FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
              AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
              LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
              OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
              THE SOFTWARE.

'''

__version__ = (1, 1)

# stdlib
import os
import sys
import csv
import abc
import time
import json
import base64
import string
import logging

# timezone data
from tz import timezone
from tz import _TZNAMES
from tz import _TZSHORT
from tz import _TIMEZONES

# datetime tools
from datetime import date
from datetime import tzinfo
from datetime import datetime
from datetime import timedelta


## Globals

# internal constants
_CSV_FILE, _CSV_PARAMS = 'importer-events.csv', {'escapechar': '\\', 'delimiter': ',', 'quoting': csv.QUOTE_NONE}

# keen-related constants
_KEEN_DATE_FMT, _KEEN_DATE_FMT_ISO = '%Y-%m-%dT%H:%M:%S', '%Y-%m-%dT%H:%M:%S%z'

# mixpanel-related constants
_MIXPANEL_DATE_FMT = '%Y-%m-%d'
_MIXPANEL_API, _MIXPANEL_EXPORT = 'http://mixpanel.com/api', 'https://data.mixpanel.com/api'

# `_TO_TIMESTAMP`: converts a python datetime into a integer timestamp
_TO_TIMESTAMP = lambda dt: int(time.mktime(dt.timetuple()))

# `_PROVIDER`: converts a provider name ("keen") to a sentinel for that provider (`KEEN`)
_PROVIDER = lambda name: getattr(Providers, name.upper())  # return ``KEEN`` for ``keen``
_PROVIDER_IMPL = lambda sentinel: globals()[''.join(sentinel.capitalize())]  # return :py:class:`Keen` for ``KEEN``

# `_PRETTY_DATE`: output a string repr for a date that's easy on the eyes
_PRETTY_DATE = lambda _date: _date.strftime('%a %b %d, %Y')

# `_[TO/FROM]_INTERNAL_DT`: converts between a serialized and in-memory representation of an internal datetime object
_INTERNAL_DATE_FMT = '%Y-%m-%d::%H:%M:%S'
_TO_INTERNAL_DT = lambda dt: dt.astimezone(timezone('UTC')).strftime(_INTERNAL_DATE_FMT)
_FROM_INTERNAL_DT = lambda spec: timezone('UTC').localize(datetime.strptime(spec, _INTERNAL_DATE_FMT))

# `_[TO/FROM]_KEEN_DATE`: converts to/from ISO8601 format and Python `date`
_TO_KEEN_DATE = lambda _date: ''.join((_date.strftime(_KEEN_DATE_FMT_ISO), 'Z' if not _date.tzinfo else ''))
_FROM_KEEN_DATE = lambda spec: timezone('UTC').localize(datetime.strptime('-'.join(spec.replace('Z', '').split('-')[0:3]).split('+')[0], _KEEN_DATE_FMT))

# `_[TO|FROM]_MIXPANEL_DATE`: converts to/from YYYY-MM-DD format and Python `date`
_TO_MIXPANEL_DATE = lambda _date: _date.strftime(_MIXPANEL_DATE_FMT)
_FROM_MIXPANEL_DATE = lambda spec: datetime.strptime(spec, _MIXPANEL_DATE_FMT)

# `PROVIDERS`: global list of installed providers
PROVIDERS = {}


# +=+=+=+ Abstract Classes +=+=+=+ #

class Providers:

    ''' Enumerates available event providers. '''

    pass


class Provider(object):

    ''' Small wrapper object that binds config,
        an :py:class:`Uploader` class and a
        :py:class:`Downloader` class that
        (together) add a provider for ``Importer``. '''

    class __metaclass__(abc.ABCMeta):

        ''' Provider-specific metaclass that registers
            new providers globally. '''

        def __new__(cls, name, bases, properties):

            ''' Construct new provider and register
                with global dict of providers.

                :param name: Class name for construction.
                :param bases: Class bases for construction.
                :param properties: Class-level property map for construction.
                :returns: Constructed :py:class:`Provider` class. '''

            global PROVIDERS

            # construct class (careful of the ABC inheritance tree)
            klass = super(cls, cls).__new__(cls, name, bases, properties)

            # don't register root provider
            if name is 'Provider': return klass

            # register, install & return
            PROVIDERS[name] = klass
            setattr(Providers, name.upper(), name.upper())
            return PROVIDERS[name]

    bus = None  # reference upwards to ``Importer``
    lib = None  # library for this provider
    name = None  # string name of this provider
    client = None  # current client object, from ``lib`` (if applicable)
    config = None  # config from ``config.json``
    adapters = None  # adapter uploader/downloader pair

    def __init__(self, bus, name, library, adapters, **config):

        ''' Initialize this ``Provider``, given
            a provider name and uploader/downloader
            pair.

            :param bus: Reference upwards to the main
            :py:class:`Importer` object.

            :param name: String provider name from
            :py:class:`Provider`, like ``'KEEN'``
            or ``'MIXPANEL'``.

            :param adapters: Tupled pair of a
            :py:class:`Uploader` object and matching
            :py:class:`Downloader` object.

            :param **config: Provider-specific config
            values found in ``config.json``.

            :returns: Nothing, as this method is a
            constructor. '''

        self.bus, self.name, self.lib, self.config, self.adapters = (
            bus,
            name,
            library,
            config,
            adapters
        )

        if not isinstance(adapters, tuple) or len(adapters) < 2:
            raise TypeError('Must provide a valid `Downloader` and `Uploader`'
                            ' tupled pair to the `Provider` constructor. Got: `%s`.' % adapters)

        self.initialize()  # dispatch provider-level initialize

    @property
    def uploader(self):

        ''' Grab and initialize this ``Provider``'s
            local :py:class:`Uploader`, if there isn't
            one already instantiated.

            :returns: An instantiated and configured
            :py:class:`Uploader`, for this ``Provider``. '''

        uploader, downloader = self.adapters
        if isinstance(uploader, type):  # does it need initializing?
            adapter = uploader(self)
            self.adapters = (adapter, downloader)
            return adapter
        return uploader  # already instantiated

    @property
    def downloader(self):

        ''' Grab and initialize this ``Provider``'s
            local :py:class:`Downloader`, if there isn't
            one already instantiated.

            :returns: An instantiated and configured
            :py:class:`Downloader`, for this ``Provider``. '''

        uploader, downloader = self.adapters
        if isinstance(downloader, type):  # does it need initializing?
            adapter = downloader(self)
            self.adapters = (uploader, adapter)
            return adapter
        return downloader  # already instantiated

    @property
    def logging(self):

        ''' Build a child logger to the root ``importer``
            logger that is specific to this ``Provider``.

            :returns: Customized :py:mod:`logging.Logger`
            object. '''

        if self.bus and hasattr(self.bus, 'logging'):  # pragma: nocover
            return self.bus.logging.getChild(self.name.capitalize())
        return logging

    def __repr__(self):

        ''' Generate a nice string representation
            of this provider.

            :returns: A string like ``'Provider(KEEN)'``. '''

        return 'Provider(%s)' % self.name

    @abc.abstractmethod
    def initialize(self):  # pragma: nocover

        ''' Initialization hook for :py:class:`Provider`
            implementors. Dispatched by :py:meth:`__init__`
            upon provider construction.

            :raises NotImplementedError: If the root abstract
            method is called directly on the ABC-enforced
            class.

            :returns: Nothing, as this is a constructor-like
            method. '''

        raise NotImplementedError('`Provider.initialize` is abstract'
                                  ' and cannot be dispatched directly.')


class Adapter(object):

    ''' Abstract class that specifies basic
        methods that both :py:class:`Uploader`
        and :py:class:`Downloader` share. '''

    utc = timezone('UTC')

    # == Encapsulated ABC Enforcer == #
    class __metaclass__(abc.ABCMeta):

        ''' Encapsulated adapter metaclass that
            enforces ABC semantics and prevents
            :py:class:`Adapter` children that are
            not :py:class:`Uploader` or
            :py:class:`Downloader`. '''

        def __new__(cls, name, bases, prop_map):

            ''' Factory a new :py:class:`Adapter` class,
                by yielding to :py:mod:`abc` if it is
                approved for implementation. '''

            prop_map['__name__'] = name  # set name
            prop_map['__bases__'] = bases  # set bases

            # allow this class (obviously)
            if (name == 'Adapter' and bases == (object,)):

                # overwrite class name to hide metaclass magic
                cls.__name__ = name

                # yield to ABCMeta, but run in this class' context
                return type.__new__(cls, name, bases, prop_map)

            # disallow non (downloaders || uploaders)
            if name in frozenset(('Uploader', 'Downloader')):
                if len(bases) == 1 and bases[0] is Adapter:
                    # it's a valid adapter
                    return super(Adapter.__metaclass__, cls).__new__(cls, name, bases, prop_map)

            # implementors of `Uploader` and `Downloader` are obviously allowed
            if Uploader in bases or Downloader in bases:
                return super(Adapter.__metaclass__, cls).__new__(cls, name, bases, prop_map)

            # only :py:class:`Uploader` and :py:class:`Downloader` can implement :py:class:`Adapter`
            raise RuntimeError('`Adapter` does not support implementations other than'
                               ' `Downloader` and `Uploader`. If you would like to extend'
                               ' `importer`, you should implement a `Provider` with an'
                               ' `Uploader`/`Downloader` pair. See docs for more info.')


    # == Class Members == #
    provider = None  # bridge up to this adapter's provider

    def __new__(cls, *args, **kwargs):

        ''' Enforce that abstract :py:class:`Adapter`
            implementors cannot be instantiated directly,
            as they are abstract and define no real
            functionality.

            This method passes everything on to
            :py:meth:`__init__`, assuming the class being
            constructed passes our tests.

            :param args: Positional construction arguments.
            :param kwargs: Keyword-based construction arguments. '''

        if cls.__name__ in frozenset(('Adapter', 'Uploader', 'Downloader')):
            raise TypeError('Cannot instantiate `%s` directly, as it is abstract.' % cls.__name__)
        return super(Adapter, cls).__new__(cls, *args, **kwargs)

    def __init__(self, provider=None):

        ''' Initialize this :py:class:`Adapter`.
            Takes a Python library to attach locally
            and a set of config handed in from the
            local :file:`config.json`.

            :param provider: Parent :py:class:`Provider`
            encapsulating this ``Adapter``. Allows access
            to ``Provider``-specific config and a ref
            to this ``Adapter``'s library.

            :returns: Nothing, as this method is a
            constructor. '''

        # must be instantiated from a provider, not directly
        if not provider:
            raise RuntimeError('Cannot instantiate an `Adapter` or direct subclass'
                               ' directly. `Adapter` classes must be used from a'
                               ' valid `Provider`.')

        # set provider and propagate to `self.initialize`, if present
        self.provider = provider
        if hasattr(self, 'initialize'):  # pragma: nocover
            self.initialize()

    @property
    def logging(self):

        ''' Build a child logger to the root ``importer``
            logger that is specific to this ``Provider``.

            :returns: Customized :py:mod:`logging.Logger`
            object. '''

        if self.provider and hasattr(self.provider, 'logging') and not isinstance(self.provider.logging, type(os)):  # pragma: nocover
            # return child logger to provider, like `Uploader` so it ends up as `keen.Uploader`
            return self.provider.logging.getChild(self.__class__.__name__.replace(self.provider.__class__.__name__, ''))
        return logging

    @abc.abstractmethod
    def transform(self, data):  # pragma: nocover

        ''' Hook for :py:class:`Adapter` implementors that
            allows transformations to be applied to data
            moving up (in the case of an :py:class:`Uploader`)
            or down (in the case of a :py:class:`Downloader`).

            The :py:meth:`transform` method is called once per event
            moving through the :py:class:`Adapter`. Think of it as a
            ``map()``-type hook, as the return value is taken **in place**
            of the original data.

            If the adapter means not to implement :py:meth:`transform`,
            it can return the special value ``NotImplemented``, which
            indicates that transformation is not required. If
            ``NotImplemented`` is not the return value, the result is
            taken as the transformed data.

            :param data: Raw data to transform.

            :raises: ``NotImplementedError``, if the root abstract
            method is called, via ``super()`` or directly.

            :returns: Transformed data, or ``NotImplemented``, if
            this class does not implement transformations. '''

        raise NotImplementedError('`Provider.transform` is abstract'
                                  ' and cannot be dispatched directly.')


class Uploader(Adapter):

    ''' Abstract class for an Uploader object,
        which manages the process of uploading
        events to a particular service. '''

    def pre_process(self, data):

        ''' Pre-process hook that implements the transform
            flow for data staged for upload. Called once
            for each event that is due to be uploaded.

            :param data: Raw data that would be uploaded, usually
            from the buffered CSV file of downloaded data.

            :raises: Nothing.

            :returns: Transformed data, from :py:meth:`transform`,
            or the unmodified data if no transform was supported/
            accepted. '''

        transformed = self.transform(data)
        return transformed if transformed is not NotImplemented else data

    @abc.abstractmethod
    def upload(self, kind, data, timestamp=None):  # pragma: nocover

        ''' Abstract method specifying the existence
            (and interface) to ``upload`` some event
            data. Must be overridden by a subclass.

            :param kind: Event kind name to upload.

            :param data: Event data to upload.

            :param timestamp: Timestamp for the new
            event. ``datetime.now()`` is taken if
            ``None`` is passed.

            :raises NotImplementedError: Always, as
            this method is abstract. '''

        raise NotImplementedError()

    @abc.abstractmethod
    def commit(self):  # pragma: nocover

        ''' Event commit hook, called at the end of the
            upload flow to indicate the stream of events
            has been exhausted and is ready to send.

            :py:class:`Uploader` implementors must define
            this method.

            :returns: The count of buffered events that
            are due to be uploaded, as an ``int``.'''

        raise NotImplementedError()

    def transform(self, data):

        ''' Data transformation hook. Left unimplemented
            by default.

            :param data: Pre-transform data to perform
            our work on.

            :raises: Nothing.

            :returns: ``NotImplemented``, a special value
            specifying that this method is unimplemented
            by default. '''

        return NotImplemented


class Downloader(Adapter):

    ''' Abstract class for a Downloader object.
        Responsible for managing the process of
        downloading events from a particular
        service. '''

    def post_process(self, data):

        ''' Post-process data that has just been downloaded,
            to cleanse/tune the data for compatibility or
            better value to ``Importer``.

            :param data: Raw data that has recently been
            downloaded and is awaiting transformation.

            :raises: Nothing.

            :returns: Transformed data, from :py:meth:`transform`,
            or the unmodified data if no transform was supported/
            accepted. '''

        transformed = self.transform(data)
        return transformed if transformed is not NotImplemented else data

    @abc.abstractmethod
    def download(self, begin, end, kinds):  # pragma: nocover

        ''' Abstract method specifying the existence
            (and interface) to ``download`` some event
            data. Must be overridden by a subclass.

            :param begin: Begin date for event data
            to download.

            :param end: End date for event data to
            download.

            :param kinds: Event kind names to
            download.

            :raises NotImplementedError: Always, as
            this method is abstract. '''

        raise NotImplementedError()

    def transform(self, data):

        ''' Data transformation hook. Left unimplemented
            by default.

            :param data: Pre-transform data to perform
            our work on.

            :raises: Nothing.

            :returns: ``NotImplemented``, a special value
            specifying that this method is unimplemented
            by default. '''

        return NotImplemented


# +=+=+=+ Keen Classes +=+=+=+ #

class Keen(Provider):

    ''' Handles operations specific to ``Keen``. '''

    def initialize(self):

        ''' Initialize this ``Keen`` provider
            instance.

            :returns: Nothing, as this method
            is a delegated constructor. '''

        self.lib.project_id, self.lib.write_key, self.lib.read_key = (
            self.config.get('project_id'),
            self.config.get('write_key'),
            self.config.get('read_key')
        )

        if self.config.get('endpoint'):  # custom endpoint?
            self.lib.api.KeenApi.base_url = self.config.get('endpoint')

    def _set_config(self, name, value):

        ''' Set an internal config value, such
            as ``project_id`` or ``write_key``/
            ``read_key``.

            Just proxies access to module-level
            globals in ``Keen``, or wherever
            config exists for ``Keen``.

            :param name: Name of the config value
            to set.

            :param value: Value to set the config
            entry to.

            :raises AttributeError: If the config
            key is not in the list of valid config
            items, which is ``project_id``,
            ``write_key`` and ``read_key``.

            :returns: ``self``, for chainability. '''

        # only set globals that are allowed
        if name not in frozenset(('project_id', 'read_key', 'write_key')):
            raise ValueError('Cannot set unknown Keen config value: `%s`.' % name)

        # it's valid, set it and return `self`
        self.config[name] = value
        setattr(self.lib, name, value)
        return self

    # `_[get/set]_read_key`: small get/set methods to manage a read key
    _get_read_key, _set_read_key = (lambda self: self.lib.read_key), (lambda self, value: self._set_config('read_key', value))

    # `_[get/set]_write_key`: small get/set methods to manage a write key
    _get_write_key, _set_write_key = (lambda self: self.lib.write_key), (lambda self, value: self._set_config('write_key', value))

    # `_[get/set]_project_id`: small get/set methods to manage a project ID
    _get_project_id, _set_project_id = (lambda self: self.lib.project_id), (lambda self, value: self._set_config('project_id', value))

    # `read_key/write_key/project_id`: get/set proxies for config values
    read_key, write_key, project_id = property(_get_read_key, _set_read_key), property(_get_write_key, _set_write_key), property(_get_project_id, _set_project_id)


class KeenUploader(Uploader):

    ''' Uploads CSV-based event data to Keen.
        Implements :py:class:`Uploader`. '''

    events_by_kind = {}
    timezone = timezone('UTC')

    def upload(self, kind, data, timestamp=None):

        ''' Upload some event data to ``Keen``.
            Implements the abstract method
            :py:meth:`Uploader.upload`.

            :param kind: Kind name of event to
            upload to ``Keen``.

            :param data: Data for event to upload
            to ``Keen``.

            :param timestamp: Timestamp to set for
            new event. ``datetime.now()`` is taken
            if ``None`` is passed.

            :returns: Result of the low-level Keen
            ``add_event`` call. '''

        if timestamp and not isinstance(timestamp, datetime):
            raise RuntimeError('Timestamps passed to `upload` must be `None` or a Python datetime.')

        # provision kind bucket if it's not there
        if kind not in self.events_by_kind:
            self.events_by_kind[kind] = []

        # decode event JSON
        deserialized = dict(((k.replace('$', ''), v) for k, v in json.loads(data).iteritems()))

        # assign timestamp, if available #@TODO(sgammon): fix timestamps
        try:
            if 'keen' not in deserialized:
                if timestamp:
                    if not timestamp.tzinfo:
                        # timestamp should come in as UTC, make sure it is
                        self.timezone.localize(timestamp)
                else:
                    timestamp = self.timezone.localize(datetime.now())

                deserialized['keen'] = {'timestamp': _TO_KEEN_DATE(timestamp)}

        except TypeError:
            self.logging.critical('Failed to decode timestamp for event!')

        # buffer and return
        return self.events_by_kind[kind].append(deserialized)

    def commit(self):

        ''' Flush the local event buffer to Keen.

            :returns: Result of the low-level Keen call. '''

        # add events via client
        count, result = (
            sum(map(len, self.events_by_kind.itervalues())),  # grab summed length of all buffered events
            self.provider.lib.add_events(self.events_by_kind)  # add the events to Keen (this results in a web request)
        )

        # return count of events we sent
        return count


class KeenDownloader(Downloader):

    ''' Downloads event data from Keen into CSV.
        Implements :py:class:`Downloader`. '''

    def download(self, begin, end, kinds):

        ''' Download some event data from ``Keen``.
            Implements the abstract method
            :py:meth:`Downloader.download`.

            :param begin: Begin date for events to
            download from ``Keen``. Accepts a Python
            :py:class:`datetime.datetime`.

            :param end: End date for events to
            download from ``Keen``. Accepts a Python
            :py:class:`datetime.datetime`.

            :param kinds: Event kind names to
            to download from ``Keen``. Accepts any
            nonduplicate iterable set of ``basestring``s.

            :raises NotImplementedError: Always, as
            this method is currently not implemented.

            :returns: ``yields`` the event ``kind``,
            the ``deserialized`` event, and the event's
            decoded ``timestamp``, in that order, packed
            into a tuple. '''

        _buffered_kinds = []
        for kind in kinds:

            # `extract` for kind, deserialize and yield
            for deserialized in self.provider.lib.extraction(kind, {'start': _TO_KEEN_DATE(begin), 'end': _TO_KEEN_DATE(end)}):

                # pull out timestamp and yield
                yield (kind, deserialized, _FROM_KEEN_DATE(deserialized.get('keen', {}).get('timestamp')))  # this will come out as UTC already (yay!)


# +=+=+=+ Mixpanel Classes +=+=+=+ #

class Mixpanel(Provider):

    ''' Handles operations specific to ``Mixpanel``. '''

    gmt = timezone('GMT')  # static ``GMT`` timezone because why not
    timezone = None  # account timezone :py:class:`tzinfo` descendant

    def initialize(self):

        ''' Initialize this ``Mixpanel`` provider
            instance.

            :returns: Nothing, as this method
            is a delegated constructor. '''

        # resolve timezone info and factory class
        self.timezone = timezone(self.config.get('timezone', 'UTC'))

        # setup API client with keys and proper endpoint
        self.client = self.lib.Mixpanel(self.config.get('api_key'), self.config.get('api_secret'))
        self.client.ENDPOINT = _MIXPANEL_API

    def __toggle_export__(self, type=None, value=None, traceback=None):

        ''' Enter or exit "Export Mode," which switches
            the Mixpanel API endpoint, since it is
            different for the Export API.

            :param type:
            :param value:
            :param traceback:

            :returns: ``self``, for method
            chainability. '''

        # resolve new endpoint and toggle
        self.client.ENDPOINT = _MIXPANEL_EXPORT if (self.client.ENDPOINT == _MIXPANEL_API) else _MIXPANEL_API
        self.logging.debug('Switched Mixpanel endpoint to "%s".' % self.client.ENDPOINT)

        if traceback: raise  # always re-raise exceptions (we can't stop, this is bat country)
        
        return self

    # `__enter__`/`__exit__`: co-opt Python's `with` for endpoint switching
    __enter__ = __exit__ = lambda self, *args: self.__toggle_export__(*args)


class MixpanelUploader(Uploader):

    ''' Uploads CSV-based event data to Mixpanel.
        Implements :py:class:`Uploader`. '''

    events_by_kind = {}

    def upload(self, kind, data, timestamp=None):

        ''' Upload some event data to ``Mixpanel``.
            Implements the abstract method
            :py:meth:`Uploader.upload`.

            :param kind: Kind name of the event
            to upload to ``Mixpanel``.

            :param data: Data for the event to
            upload to ``Mixpanel``.

            :param timestamp: Timestamp to set
            for the new event. ``datetime.now()``
            is taken if ``None`` is passed.

            :raises NotImplementedError:
            :returns: '''

        # allocate kind buffer, if it's not there already
        if kind not in self.events_by_kind:
            self.events_by_kind[kind] = []

        # deserialize event from JSON, replacing dollar signs in keys as we go
        deserialized = dict(((k.replace('$', ''), v) for k, v in json.loads(data).iteritems()))

        if 'time' in deserialized:
            deserialized['original_time'] = deserialized['time']
            deserialized['time'] = int(time.mktime(timestamp.astimezone(self.provider.gmt).timetuple()))

        #@TODO(sgammon): MUST CONVERT TO GMT. WHY????????? WTF JUST WHY

        # buffer event with the rest of its kind
        return self.events_by_kind[kind].append(deserialized)

    def commit(self):

        ''' Commit buffered events to ``Mixpanel``.
            Takes buffered events from
            :py:attr:`events_by_kind` and uploads
            in batches.

            :returns: Total count of events uploaded
            to ``Mixpanel``, across all batches. '''

        # count buffered events
        count = sum(map(len, self.events_by_kind.itervalues()))

        for kind in self.events_by_kind:

            for event in self.events_by_kind[kind]:

                # kick off uploader job
                result = self.provider.client.request(['import'], {
                    'ip': 0,  # we're not the event source, just a relay
                    'data': base64.b64encode(json.dumps(event))
                })

        return count


class MixpanelDownloader(Downloader):

    ''' Downloads event data from Mixpanel into CSV.
        Implements :py:class:`Downloader`. '''

    def download(self, begin, end, kinds):

        ''' Download some event data from ``Mixpanel``.
            Implements the abstract method
            :py:meth:`Downloader.download`.

            :param begin: Begin date for event range
            to download from ``Mixpanel``.

            :param end: End date for event range to
            download from ``Mixpanel``.

            :param kinds: List of kind names for event
            types to download.

            :raises RuntimeError: Re-raises internal
            ``Mixpanel`` exceptions as ``RuntimeError``.

            :returns: Yields each event as it is received,
            in the format ``tuple(kind, data, timestamp)``.
            The ``timestamp`` tuple element is a Python
            ``datetime``, localized to the Mixpanel account's
            timezone (set in ``config.json`` at
            ``mixpanel['timezone'])``, and converted to ``UTC``. '''

        with self.provider as mixpanel:  # enter `export` mode, which switches endpoints

            # export events for each kind in the date range
            result = mixpanel.client.request(['export'], {
                'from_date': _TO_MIXPANEL_DATE(begin),
                'to_date': _TO_MIXPANEL_DATE(end),
                'event': kinds
            })

            # check for mixpanel API errors
            if 'error' in result:  # pragma: nocover

                # @TODO(sgammon): proper mixpanel error handling
                if self.provider.bus.debug:
                    import pdb; pdb.set_trace()

                raise RuntimeError(' '.join(('Mixpanel:', result['error'])))

            else:  # no error ocurred: parse newline-separated JSON events
                event_count, error_count, ignore_errors = 0, 0, False
                for event in filter(lambda x: x, result.split("\n")):  # (subtly filter out falsy empty strings)
                    
                    event_count += 1
                    self.logging.debug('Got raw event: "%s".' % event)

                    try:

                        # try to deserialize
                        deserialized = json.loads(event)

                    except Exception as e:  # pragma: nocover

                        self.logging.debug('Encountered JSON exception: "%s".' % str(e))
                        self.logging.error('Failed to decode JSON from blob.')

                        # prompt for what to do from here
                        if (not ignore_errors) and self.provider.bus.prompt("\n\n`Importer` encountered an issue decoding the following blob:\n%s\n" % event, "Ignore errors?"):
                            ignore_errors = True

                        # fail if we've tried more than 5 times and every time has failed
                        error_count += 1
                        if (event_count > 5) and error_count == event_count:
                            raise RuntimeError(' '.join(('Mixpanel:', 'too many errors.')))

                    else:

                        # convert to python datetime
                        ts = deserialized['properties']['time']
                        if isinstance(ts, int):
                            ts = datetime.fromtimestamp(ts)

                        # yield event to caller
                        yield (deserialized['event'], deserialized['properties'], self.provider.timezone.localize(ts).astimezone(self.utc))


# +=+=+=+ Importer +=+=+=+ #

class Importer(object):

    ''' Logic to stitch :py:class:`Uploader` and
        :py:class:`Downloader` classes together
        into a handy CLI tool. '''

    # == Public Members == #
    cli = False  # are we running via CLI?
    yes = True  # assume `yes` to all prompts
    debug = False  # whether debug mode is active
    quiet = False  # whether quiet mode is active
    argset = None  # arguments passed to us via command line
    log_level = logging.INFO  # default logging level
    log_format = "%(levelname)s [importer]: %(message)s"  # format string for logs

    # == Private Members == #
    __logger = None  # dedicated logger
    __buffer = None  # buffer used for writing / reading events
    __config = None  # interpreted config object (`config.json`)
    __providers = None  # references to current provider adapters

    def __init__(self, keen=None, mixpanel=None, strict=False):

        ''' Initialize this :py:class:`Importer` run.
            Take dependent libraries injected via
            parameters to this method and prepare to
            run the tool.

            :param keen: ``Keen`` API client library to
            use for ``Keen`` operations. If ``None`` is
            provided, ``Importer`` attemps to load ``Keen``
            itself with a regular import.

            :param mixpanel: ``Mixpanel`` API client
            library to use for ``Mixpanel`` operations.
            If ``None`` is provided, ``Importer`` attempts
            to load ``Mixpanel`` itself with a regular
            import.

            :param strict: Fail if we aren't provided with
            ``keen``/``mixpanel`` libraries, and they are
            not immediately importable at expected default
            paths.

            :raises ImportError: Re-raises encountered library
            import issues, if libraries are not provided via
            the regular constructor params.

            :returns: Nothing, as this method is
            a constructor. '''

        self.__local_logging__ = not (not keen and not mixpanel)  # direct instantiation probably, turn off local logging

        # if we're not handed libraries, fix `sys.path` because we're probably being imported
        if not keen and not mixpanel:  # @TODO(sgammon): why did I ever do this?
            self.fix_path()

        if not keen:

            try:

                if 'keen' not in sys.modules:
                    import keen
                else:  # if it's an explicit `None` it's coming from the test framework, fail it
                    if sys.modules['keen'] is None:
                        raise ImportError('Failed to resolve library `Keen`.')
                    keen = sys.modules['keen']

            except ImportError:

                self.logging.critical('Failed to resolve `Keen` library. Integration with Keen may not work.')
                if strict:
                    self.logging.critical('Strict mode is active. Failing because we can\'t find `keen`.')
                    raise

        if not mixpanel:

            try:

                if 'mixpanel' not in sys.modules:
                    import mixpanel
                else:  # if it's an explicit `None` it's coming from the test framework, fail it
                    if sys.modules['mixpanel'] is None:
                        raise ImportError('Failed to resolve library `Mixpanel`.')
                    mixpanel = sys.modules['mixpanel']

            except ImportError:

                self.logging.critical('Failed to resolve `Mixpanel` library. Integration with Mixpanel may not work.')
                if strict:
                    self.logging.critical('Strict mode is active. Failing because we can\'t find `mixpanel`.')
                    raise

        # bind libraries locally
        self.__providers = {
            Providers.KEEN: (keen, Keen, KeenUploader, KeenDownloader) if keen else None,
            Providers.MIXPANEL: (mixpanel, Mixpanel, MixpanelUploader, MixpanelDownloader) if mixpanel else None
        }

    @property
    def config(self):

        ''' Return loaded configuration for the current
            :py:class:`Importer` run. This may involve
            parsing the :file:`config.json` if it hasn't
            been already.

            :returns: Interpreted config blob from
            ``config.json`` file, or ``{'debug': True}``. '''

        if not self.__config:

            # resolve configuration path
            if self.argset and hasattr(self.argset, 'config_file') and self.argset.config_file:
                config_path = self.argset.config_file
            else:
                config_path = os.path.join(os.path.dirname(__file__), 'config.json')

            # try loading config
            try:
                self.logging.debug("Loading 'config.json'...")
                with open(config_path, 'r') as config_handle:
                    self.__config = json.loads(config_handle.read())

            except (IOError, OSError, ValueError, TypeError) as e:  # pragma: nocover

                # raise warning and continue
                self.__config = {'importer': {}, 'keen': {}, 'mixpanel': {}, 'debug': True}
                self.logging.warning('Failed to parse `config.json` file. Falling back to default config. '
                                     'Got exception: "%s".' % str(e))

        return self.__config

    # `mixconfig`: shortcut to grab `importer` config
    mixconfig = property(lambda self: self.config.get('importer', {'debug': True}))

    @property
    def logging(self):

        ''' Local proxy to Python logging for the
            :py:class:`Importer` tool.

            :returns: Configured :py:mod:`logging`
            channel. '''

        if self.__local_logging__:
            if not self.__logger:
                self.__logger = logging.getLogger().getChild('importer')
            return self.__logger  # @TODO(sgammon): file-based logging
        return logging

    @staticmethod
    def fix_path():

        ''' Fix the system path, if we're running
            :py:class:`Importer` directly instead
            of via the CLI, which fixes the path
            for us.

            :returns: :py:attr:`sys.path`, because
            why not. '''

        prdir = os.path.dirname(__file__)  # get project directory and lib directory
        libdir = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'lib')

        for i in (prdir, libdir):
            if i not in sys.path:
                sys.path.insert(0, i)

    def say(self, message):  # pragma: nocover

        ''' Quick shortcut for info-type messages
            that are procedural and not meant for
            logging.

            :param message: String message to print.
            :returns: ``self``, for chainability. '''

        print '[importer]: %s' % message
        return self

    def prompt(self, text, prompt='Continue?'):  # pragma: nocover

        ''' Prompt the user if we should continue
            operating. If ``yes`` mode is on, skip
            the prompt and assume a truthy value.

            :param text: String to output to the
            user before prompting.

            :param prompt: Question to prompt the
            user with after printing ``text``.
            Defaults to "Continue?".

            :returns: Bool indicating whether the
            user said we should continue, or
            ``True`` in every case if ``self.yes``
            is activated. '''

        if text and not self.quiet: print text
        return self.yes or (
            raw_input('%s [y/n]: ' % prompt) in ('y', 'yes', 'yup', 'mmk', 'go ahead', '1')
        )

    def provider(self, code, instance=True, _all=False):

        ''' Provide a small wrapper and prepared
            :py:class:`Uploader`/:py:class:`Downloader`
            pair for a certain provider.

            :param code: Provider code, mapped at
            :py:class:`Provider`.

            :param instance: Bool flag indicating whether
            a full instance is requested, or we just need
            the ``provider``'s class/construction params.

            :param _all: Bool flag indicating whether the
            construction params that *would be* or *were*
            used (depending on ``instance``) be returned
            in the response, as the first item in a two-
            item ``tuple``.

            :raises NotImplementedError: If the requested
            provider ``code`` cannot be resolved to a valid
            provider. Checking provider support is easy -
            simply use the ``in`` operator with the provider
            name and the ``Importer`` object itself: ``'keen' in Importer``.

            :raises RuntimeError: Re-raises exceptions
            that occur during construction of the
            :py:class:`Provider` object.

            :returns: Instance of :py:class:`Provider`,
            prepared with a :py:class:`Uploader` and
            :py:class:`Downloader` and all relevant
            config. '''

        try:
            # covert to provider code
            code = _PROVIDER(code)

        except AttributeError:  # nothing we've ever heard of before
            raise NotImplementedError('Requested provider `%s` is not supported.' % code)

        if code not in self:  # pragma: nocover
            raise NotImplementedError('Requested provider `%s` is not supported '
                                      'or not implemented.' % code)

        # grab library and up/down classes
        self.logging.info('Loading provider: "%s".' % code)
        library, provider, uploader, downloader = self.__providers[code]

        if instance:
            p = provider(
                self,
                code,
                library,
                (uploader, downloader),
                **self.config.get(code.lower(), {'debug': True})
            )
            return p
        if _all:
            return (library, provider, uploader, downloader), provider
        return provider

    def __contains__(self, provider):

        ''' Check ``Importer`` for support for the given
            provider name, which is resolved through
            :py:func:`_PROVIDER` and checked for support.

            Following the ``__contains__`` protocol, ``True``
            is returned if ``provider`` is supported, else
            ``False``.

            :param provider: Name (``str``) of the provider
            we wish to check support for. '''

        return _PROVIDER(provider) in self.__providers

    def execute(self, begin, end, **arguments):

        ''' Execute :py:class:`Importer`, performing
            the requested operations as interpreted
            via :file:`config.json` and the provided
            command-line arguments.

            :param begin: Python py:class:`datetime.date`
            object representing the start of the range
            of events to transfer.

            :param end: Python py:class:`datetime.date`
            object representing the end of the range
            of events to transfer.

            :param **arguments: Args for ``importer``,
            usually passed in via the command line.
            Takes the same keyword arguments as the
            CLI interface.

            :returns: Bool indicating whether this
            ``Importer`` operation was successful.
            ``True`` indicates success. '''

        self.logging.debug('Got arguments: %s' % str(arguments))
        self.logging.debug('Got range: BEGIN(%s), END(%s)' % (begin.isoformat(), end.isoformat()))

        self.yes, self.debug, self.quiet, self.verbose = (
            arguments.get('yes', False),
            arguments.get('debug', False),
            arguments.get('quiet', False),
            arguments.get('verbose', False)
        )

        # grab kinds, to/from providers
        kinds, source, target, reset_ts = [arguments[x] for x in ('kinds', 'from', 'to', 'reset_ts')]
        
        # resolve provider object
        source, target = self.provider(source), self.provider(target)
        self.logging.debug('Providers: SOURCE(%s) -> TARGET(%s)' % (source, target))

        # unless we were told to 'yes' all prompts, confirm what we're doing
        if not self.yes and not self.prompt("""

    Importer plans to:
        -- Transfer event data from %s to %s
        -- That occurred between %s and %s
        -- %s existing timestamps
        -- For the following event types: %s

    """ % (

            source.name.capitalize(),
            target.name.capitalize(),
            _PRETTY_DATE(begin),
            _PRETTY_DATE(end),
            'Reset' if reset_ts else 'Preserve',
            "\n" + "\n".join(map(lambda x: "\t\t-- %s" % x, kinds))

        )):  # pragma: nocover

            return True  # exit without error

        # calculate buffer file path
        download_buffer_path = os.path.join(os.path.dirname(__file__), _CSV_FILE)
        self.logging.debug('Buffering downloaded data to "%s"...' % download_buffer_path)

        # open CSV and set up download buffer file
        event_count = 0
        with open(download_buffer_path, 'wb') as download_buffer:

            # get CSV writer
            writer = csv.writer(download_buffer, **_CSV_PARAMS)

            # fetch source events
            self.say('Beginning event data download...')
            for kind, data, timestamp in source.downloader.download(begin, end, kinds):

                event_count += 1

                # log about it
                self.logging.info('Downloaded "%s" (event #%s @ %s bytes)...' % (kind, event_count, sys.getsizeof(data)))
                self.logging.debug('Got event data: "%s".' % (str(data)))

                # write to our buffer file
                writer.writerow((kind, _TO_INTERNAL_DT(timestamp), json.dumps(source.downloader.post_process(data))))

            self.logging.info('Download complete.')

        if self.cli:
            time.sleep(2)  # give the user a chance to catch up

        if event_count is 0:  # pragma: nocover
            self.say('Total events collected: 0.')
            self.say('No events to transfer. Exiting.')
            return True

        _msg_context = (str(event_count), source.name.capitalize(), target.name.capitalize())
        if (not self.yes) and not self.prompt("\n\nCollected %s events from %s.\nReady to upload to %s.\n" % _msg_context):  # pragma: nocover
            self.logging.info('Exiting on user cancel.')
            return True

        # open CSV in read mode and start upload
        event_count = 0
        with open(download_buffer_path, 'rb') as upload_buffer:

            # for each event, upload to target
            self.say('Beginning event data upload...')
            for event in csv.reader(upload_buffer, **_CSV_PARAMS):

                event_count += 1

                # extract info
                kind, timestamp, data = event

                self.logging.info('Uploading "%s" (event #%s)...' % (kind, event_count))
                self.logging.debug('Uploading event data: "%s".' % event)

                target.uploader.upload(kind, target.uploader.pre_process(data), _FROM_INTERNAL_DT(timestamp) if not reset_ts else None)  # extract row data and upload

            # dispatch commit(), if available
            try:
                self.say('Committing events...')
                target.uploader.commit()
            except NotImplementedError:  # pragma: nocover
                self.logging.critical('Failed to commit events.')

            self.say('Upload complete.')
            self.logging.info('Transferred %s events to %s.' % (str(event_count), target.name.capitalize()))

        if self.cli:
            time.sleep(2)  # give the user a chance to catch up

        if self.yes or self.prompt("\n\n`Importer` has finished transferring %s events\n "
                           "from %s to %s. The transferred data is currently\n "
                           "buffered in a file called '%s' - would you like to\n "
                           "keep it as a backup?" % (event_count, source.name.capitalize(), target.name.capitalize(), _CSV_FILE), "Keep file?"):

            try:
                os.remove(download_buffer_path)  # delete the buffer file if requested
            except:  # pragma: nocover
                self.logging.error('Failed to remove buffer file at path "%s". This is probably due to permissions.' % download_buffer_path)

        self.logging.debug('Finished work.')
        self.say('Done, exiting `Importer`.')
        return True  # exit without error

    def __call__(self, **cli):

        ''' Wrap :py:meth:`execute` for use via the
            command line. Convert exceptions into
            friendly error messages and typeconvert
            things coming in via CLI args.

            .. todo: Clean up exception handling in ``debug``
                     mode.

            .. note: Calling this method from Python is
                     possible but not recommended - if you
                     use :py:meth:`execute` instead, you can
                     pass native types and exceptions bubble
                     up properly.

            :param **cli: Command line arguments
            (``dict``) passed-in from ``__main__``'s
            :py:mod:`argparse` work.

            :returns: Unix process exit code describing
            the result of the current `Importer` run. If
            the result of :py:meth:`execute` is truthy,
            ``0`` is returned to indicate no error,
            otherwise ``1`` is returned. '''

        self.cli, self.argset, self.debug = True, cli, self.config.get('debug', False)  # setup CLI and debug mode

        try:

            # convert date strings
            try:
                begin, end = _FROM_MIXPANEL_DATE(cli['begin']), _FROM_MIXPANEL_DATE(cli['end'])
            except:  # pragma: nocover
                raise RuntimeError('Failed to parse YYYY-MM-DD begin/end dates.')  # @TODO(sgammon): proper exception handling

            # configure prompts and debug
            self.yes = cli.get('yes', False)

            # configure logging
            if cli['verbose'] > 0:  # if verbose mode is turned on...
                logging.getLogger().setLevel(logging.INFO - (10 * cli['verbose']) - 1)
                self.say('Verbose mode is active. Go more verbose with `-vv` or quiet with `-q`.')

            elif cli['quiet']:  # otherwise, if quiet mode is turned on...

                self.quiet = True  # set local quiet mode to 'on'
                logging.getLogger().setLevel(logging.CRITICAL)

            # dispatch (convert result for use as a Unix exit code)
            self.logging.info('Starting Importer...')
            return 0 if bool(self.execute(begin, end, **{

                'to': _PROVIDER(cli.get('target', self.mixconfig.get('to'))),  # source provider
                'from': _PROVIDER(cli.get('source', self.mixconfig.get('from'))),  # destination provider
                'kinds': cli.get('kinds', self.mixconfig.get('kinds')),  # event kinds to transfer
                'verbose': bool(cli.get('verbose', self.config.get('verbose', False))),  # verbose mode
                'reset_ts': bool(cli.get('reset_ts', self.config.get('reset_timestamps', False))),  # whether to reset timestamps
                'quiet': bool(cli.get('quiet', self.config.get('quiet', False))),  # quiet mode
                'yes': bool(cli.get('yes', self.config.get('yes', False)))  # yes mode

            })) else 1

        except NotImplementedError as e:  # pragma: nocover
            self.logging.error("Stubbed functionality requested.")

            # re-raise in debug mode
            if self.debug:
                import pdb; pdb.set_trace()
                raise

            print ""
            print "Some of the functionality you requested is not yet implemented:"
            print e
            print ""

            return 1

        except RuntimeError as e:  # pragma: nocover
            self.logging.error("Encountered RuntimeError.")

            # re-raise in debug mode
            #if self.debug:
            #    import pdb; pdb.set_trace()
            #    raise
            raise

            print ""
            print "An error was encountered while processing your request:"
            print e
            print ""

            return 1

        except Exception as e:  # pragma: nocover
            self.logging.critical('Unhandled exception occurred. Exiting.')

            # re-raise in debug mode
            #if self.debug:
            #    import pdb; pdb.set_trace()
            #    raise
            raise

            print ""
            print "An unhandled exception was encountered while processing your request:"
            print e
            print ""



# -*- coding: utf-8 -*-

__doc__ = '''

    `importer`:

    a handy package to transfer data
    between mixpanel.com and keen.io.

    run `make`, configure via `config.json`,
    and then run this utility to transfer
    data.

    command line flags override entries in
    `config.json`.

    :author: Sam Gammon <sam@keen.io>
    :license: This software follows the MIT (OSI-approved)
              license for open source software. A truncated
              version is included here; for full licensing
              details, see ``LICENSE.md`` in the root directory
              of the project.

              Copyright (c) 2013, Keen IO

              The above copyright notice and this permission notice shall be included in
              all copies or substantial portions of the Software.

              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
              IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
              FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
              AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
              LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
              OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
              THE SOFTWARE.

'''

__version__ = (1, 1)
__author__ = "Sam Gammon <sam@keen.io"


# stdlib
import os
import sys
import logging
import argparse

# importer
import mix
from mix import Importer

# logging
logging.basicConfig(format=Importer.log_format)

# dependencies
try:
    import keen
    import mixpanel  # try to grab libs globally

except ImportError:  # pragma: nocover

    # calculate absolute path to `importer` and patch sys.path
    if 'lib/' not in sys.path:
        [sys.path.insert(0, os.path.join(os.path.dirname(__file__), path))
            for path in ('lib/', 'lib/dist', 'lib/keen', 'lib/mixpanel')]

    try:
        import keen, mixpanel  # try to grab libs locally

    except ImportError:  # pragma: nocover

        try:  # try to grab them manually from `lib`
            if 'keen' not in globals():
                from lib import keen
            if 'mixpanel' not in globals():
                from lib.mixpanel import mixpanel

        except ImportError:
            # no dependencies: can't go on
            logging.critical('Dependencies `keen` and `mixpanel` are missing. '
                             'Please run `make` or `pip install keen mixpanel`.')

            sys.exit(1)  # exit with error


# provider choice names
_provider_choices = frozenset(map(lambda x: x.lower(), mix.PROVIDERS.iterkeys()))


## == Command line parser
parser = argparse.ArgumentParser(description=__doc__)

## == Command line flags

# begin date
parser.add_argument('begin', metavar='BEGIN', type=unicode,
                    help='start of event range to fetch (YYYY-MM-DD)')

# end date
parser.add_argument('end', metavar='END', type=unicode,
                    help='end of event range to fetch (YYYY-MM-DD)')

# config file
parser.add_argument('--config', '-c', dest='config_file', type=unicode,
                    help='config.json file with keys and settings', required=False)

# source provider
parser.add_argument('--from', '-f', dest='source', type=unicode,
                    help='provider name to get events from', choices=_provider_choices, required=False)

# destination provider
parser.add_argument('--to', '-t', dest='target', type=unicode,
                    help='provider name to send events to', choices=_provider_choices, required=False)

# event kinds
parser.add_argument('--kind', '-k', dest='kind', type=unicode, nargs='+',
                    help='event types to provide (multiple can be specified)', required=False)

# option: reset timestamps
parser.add_argument('--reset_timestamps', '-r', dest='reset_ts', action='store_true',
                    help='reset event timestamps upon upload')

# option: say `yes` to all prompts
parser.add_argument('--yes', '-y', action='store_true',
                    help='assume "yes" to all prompts')

# option: quieter output
parser.add_argument('--quiet', '-q', action='store_true',
                    help='be quiet about everything')

# option: louder output
parser.add_argument('--verbose', '-v', action='count', dest='verbose', default=0,
                    help='be loud about everything')

# option: print version and exit
parser.add_argument('--version', '-V', action='version',
                    help='print importer\'s version and exit', version='importer %s' % '.'.join(map(unicode, __version__)))


## == Importer! :)
run = lambda: sys.exit(Importer(keen=keen, mixpanel=mixpanel)(**vars(parser.parse_args())))


if __name__ == '__main__':  # pragma: nocover
    run()
